from __future__ import annotations
from hashlib import md5

from typing import Union, List, Dict, TYPE_CHECKING
from pydantic import ConfigDict, model_validator, field_validator
from prodsys.models.primitives_data import PrimitiveData
from prodsys.models.source_data import RoutingHeuristic

if TYPE_CHECKING:
    from prodsys.models.production_system_data import ProductionSystemData


class ProductData(PrimitiveData):
    """
    Class that represents product data, specifically the required processes and the allows tranport process.

    The processes describe thereby the process model that needs to be completed for the product to be finished. There are three different ways to describe the process model:

    - Sequential process model: The processes are given as a list of process IDs. The processes are executed sequentially.
    - Adjacency matrix process model: The processes are given as an adjacency matrix that describes the precedence graph of the product.
    - Edges process model: The processes are given as a list of edges. The edges describe the precedence graph of the product.

    See the examples for more insights.

    Args:
        ID (str): ID of the product. If not given, the type is used. Gets overwritten to the instance product ID, when an instance is created during simulation.
        description (str): Description of the product.
        type (str): Type of the product. If not given, the ID is used.
        processes (Union[List[str], List[List[str]], Dict[str, List[str]]]): Processes of the product. This can be a list of process IDs, a list of edges or an adjacency matrix.
        transport_process (str): Transport process of the product.
        dependency_ids (List[str]): IDs of the dependencies of the product.
        
    Examples:
        Product with sequential process model:
        ``` py
        import prodsys
        prodsys.product_data.ProductData(
            ID="Product_1",
            description="product 1",
            type="Product_1",
            processes=["P1", "P2", "P3"],
            transport_process="TP1",
        )
        ```

        Product with adjacency matrix process model:
        ``` py
        import prodsys
        prodsys.product_data.ProductData(
            ID="Product_1",
            description="Product 1",
            type="Product_1",
            processes={
                "P1": ["P2", "P3"],
                "P2": ["P3"],
                "P3": [],
            },
            transport_process="TP1",
        )
        ```

        Product with graph edges process model:
        ``` py
        import prodsys
        prodsys.product_data.ProductData(
            ID="Product_1",
            description="Product 1",
            type="Product_1",
            processes=[
                ["P1", "P2"],
                ["P1", "P3"],
                ["P2", "P4"],
                ["P3", "P4"],
            ],
            transport_process="TP1",
        )
        ```
    """

    processes: Union[Dict[str, List[str]]]
    dependency_ids: List[str] = []
    routing_heuristic: Union[RoutingHeuristic, None] = None
    becomes_consumable: bool = False
    
    @field_validator("processes", mode="before")
    def check_processes(cls, v):
        if isinstance(v, list):
            # create adjacency matrix for old API support
            process_dict = {process_id: [] for process_id in v}
            for counter, node_id in enumerate(v):
                if counter == len(v) - 1:
                    break
                process_dict[node_id].append(v[counter + 1])
            v = process_dict
        return v

    def hash(self, adapter: ProductionSystemData) -> str:
        """
        Returns a unique hash of the product considering the processes and the transport process. Can be used to compare products for equal functionality.

        Args:
            adapter (ProductionSystemAdapter): Adapter that contains the process data.

        Raises:
            ValueError: If the process or transport process is not found in the adapter.

        Returns:
            str: Hash of the product.
        """
        processes_hashes = []
        transport_process_hash = ""

        # Hash all unique processes in the adjacency matrix
        unique_process_ids = sorted(set(self.processes.keys()))
        for process_id in unique_process_ids:
            process = next((process for process in adapter.process_data if process.ID == process_id), None)
            if process is None:
                raise ValueError(
                    f"Process with ID {process_id} not found for product {self.ID}."
                )
            processes_hashes.append(process.hash(adapter))

        # TODO: add hashing for dependencies!

        for transport_process in adapter.process_data:
            if transport_process.ID == self.transport_process:
                transport_process_hash = transport_process.hash(adapter)
                break
        else:
            raise ValueError(
                f"Transport process with ID {self.transport_process} not found for product {self.ID}."
            )

        return md5(
            "".join([*processes_hashes, transport_process_hash]).encode("utf-8")
        ).hexdigest()

    @model_validator(mode="before")
    def check_type(cls, values):
        # Set type field (required by PrimitiveData)
        if "type" not in values or not values["type"]:
            values["type"] = values.get("ID", "")
        # If ID is not set, use type
        if "ID" not in values or not values["ID"]:
            values["ID"] = values.get("type", "")
        return values

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "ID": "Product_1",
                    "description": "Product with sequential process",
                    "type": "Product_1",
                    "processes": ["P1", "P2", "P3"],
                    "transport_process": "TP1",
                },
                {
                    "ID": "Product_1",
                    "description": "Process with adjacency matrix process",
                    "type": "Product_1",
                    "processes": {
                        "P1": ["P2", "P3"],
                        "P2": ["P3"],
                        "P3": [],
                    },
                    "transport_process": "TP1",
                },
                {
                    "ID": "Product_1",
                    "description": "Process with graph edges process",
                    "type": "Product_1",
                    "processes": [
                        ["P1", "P2"],
                        ["P1", "P3"],
                        ["P2", "P4"],
                        ["P3", "P4"],
                    ],
                    "transport_process": "TP1",
                },
            ]
        }
    )



import hydra
from omegaconf import DictConfig, OmegaConf

import prodsys
from prodsys.optimization.evolutionary_algorithm import run_evolutionary_algorithm
from prodsys.optimization.math_opt import MathOptimizer
from prodsys.optimization.simulated_annealing import run_simulated_annealing
from prodsys.optimization.tabu_searprodsys.adaptersrt run_tadapterch
from prodsys.simulation.runner import run_simulation
from prodsys.util.util import get_initial_solution_file_pth, prepare_save_folder


@hydra.main(version_base=None, config_path="conf", config_name="config")
def my_app(cfg: DictConfig) -> None:
    if cfg.general.mode == "simulation":
        adapter_object = adapter.ProductionSystemData()
        adapter_object.read_data(cfg.general.configuration_path)
        runner_object = run_simulation(adapter_object, cfg.simulation.simulation_length)
        if cfg.simulation.print_results:
            runner_object.print_results()
        if cfg.simulation.plot_results:
            runner_object.plot_results()
        if cfg.simulation.save_results:
            prepare_save_folder(cfg.general.save_folder)
            runner_object.save_results_as_csv(cfg.general.save_folder)
    elif cfg.general.mode == "optimization":
        if cfg.optimization.algorithm == "evolutionary":
            if (
                not "initial_solutions_folder" in cfg.general.keys()
                or not cfg.general.initial_solutions_folder
            ):
                cfg.general.save_folder += "/evolutionary"
                initial_solutions_folder = ""
            else:
                initial_solutions_folder = cfg.general.initial_solutions_folder
                cfg.general.save_folder += "/partial_evolutionary"
            prepare_save_folder(cfg.general.save_folder)
            run_evolutionary_algorithm(
                cfg.general.save_folder,
                cfg.general.configuration_path,
                cfg.general.scenario_path,
                cfg.optimization.seed,
                cfg.optimization.number_of_generations,
                cfg.optimization.population_size,
                cfg.optimization.mutation_rate,
                cfg.optimization.crossover_rate,
                cfg.optimization.number_of_processes,
                initial_solutions_folder,
            )
        elif cfg.optimization.algorithm == "simulated_annealing":
            if (
                not "initial_solutions_folder" in cfg.general.keys()
                or cfg.general.initial_solutions_folder == ""
            ):
                cfg.general.save_folder += "/simulated_annealing"
                initial_solution_file_path = ""
            else:
                cfg.general.save_folder += "/partial_simulated_annealing"
                initial_solution_file_path = get_initial_solution_file_pth(
                    cfg.general.initial_solutions_folder
                )
            prepare_save_folder(cfg.general.save_folder)
            run_simulated_annealing(
                cfg.general.save_folder,
                cfg.general.configuration_path,
                cfg.general.scenario_path,
                cfg.optimization.seed,
                cfg.optimization.Tmax,
                cfg.optimization.Tmin,
                cfg.optimization.steps,
                cfg.optimization.updates,
                initial_solution_file_path,
            )

        elif cfg.optimization.algorithm == "tabu_search":
            if (
                not "initial_solutions_folder" in cfg.general.keys()
                or not cfg.general.initial_solutions_folder
            ):
                cfg.general.save_folder += "/tabu_search"
                initial_solution_file_path = ""
            else:
                cfg.general.save_folder += "/partial_tabu_search"
                initial_solution_file_path = get_initial_solution_file_pth(
                    cfg.general.initial_solutions_folder
                )
            prepare_save_folder(cfg.general.save_folder)
            run_tabu_search(
                cfg.general.save_folder,
                cfg.general.configuration_path,
                cfg.general.scenario_path,
                cfg.optimization.seed,
                cfg.optimization.tabu_size,
                cfg.optimization.max_steps,
                cfg.optimization.max_score,
                initial_solution_file_path,
            )
        elif cfg.optimization.algorithm == "mathematical":
            cfg.general.save_folder += "/mathematical"
            prepare_save_folder(cfg.general.save_folder)
            adapter_object = adapter.ProductionSystemData()
            adapter_object.read_data(
                cfg.general.configuration_path, cfg.general.scenario_path
            )
            optimizer_object = MathOptimizer(
                adapter=adapter_object,
                optimization_time_portion=cfg.optimization.optimization_time_portion,
            )
            optimizer_object.optimize(n_solutions=cfg.optimization.n_solutions)
            if cfg.optimization.save_model:
                optimizer_object.save_model(cfg.general.save_folder)
            optimizer_object.save_results(
                save_folder=cfg.general.save_folder,
                adjusted_number_of_transport_resources=cfg.optimization.adjusted_number_of_transport_resources,
            )


if __name__ == "__main__":
    my_app()



from prodsys.optimization.evolutionary_algorithm import run_evolutionary_algorithm

seed = 23
NGEN = 9
POPULATION_SIZE = 16
N_PROCESSES = 4
NUMBER_OF_SEEDS = 1
CROSS_OVER_RATE = 0.1
MUTATION_RATE = 0.15

SAVE_FOLDER = "data/ea_results"
BASE_CONFIGURATION_FILE_PATH = (
    "examples/optimization/optimization_example/simple_primitive_example.json"
)
SCENARIO_FILE_PATH = (
    "examples/optimization/optimization_example/primitive_scenario.json"
)


if __name__ == "__main__":
    # Full optimization run
    run_evolutionary_algorithm(
        SAVE_FOLDER,
        BASE_CONFIGURATION_FILE_PATH,
        SCENARIO_FILE_PATH,
        False,
        seed,
        NGEN,
        POPULATION_SIZE,
        MUTATION_RATE,
        CROSS_OVER_RATE,
        NUMBER_OF_SEEDS,
        N_PROCESSES,
    )



"""
Tests for sink_data module.
"""

import pytest
from prodsys.models.sink_data import SinkData
from prodsys.models.product_data import ProductData
from prodsys.models.processes_data import TransportProcessData, ProductionProcessData, ProcessTypeEnum
from prodsys.models.time_model_data import FunctionTimeModelData
from prodsys.models.port_data import QueueData
from prodsys.models.production_system_data import ProductionSystemData
from prodsys.util.statistical_functions import FunctionTimeModelEnum


@pytest.fixture
def sample_time_model():
    """Create a sample time model for testing."""
    return FunctionTimeModelData(
        ID="tm1",
        description="Time model",
        distribution_function=FunctionTimeModelEnum.Normal,
        location=10.0,
        scale=2.0,
    )


@pytest.fixture
def sample_transport_time_model():
    """Create a sample transport time model for testing."""
    return FunctionTimeModelData(
        ID="tm_transport",
        description="Transport time model",
        distribution_function=FunctionTimeModelEnum.Normal,
        location=1.0,
        scale=0.1,
    )


@pytest.fixture
def sample_process(sample_time_model):
    """Create a sample production process for testing."""
    return ProductionProcessData(
        ID="P1",
        description="Process 1",
        time_model_id="tm1",
        type=ProcessTypeEnum.ProductionProcesses,
    )


@pytest.fixture
def sample_transport_process(sample_transport_time_model):
    """Create a sample transport process for testing."""
    return TransportProcessData(
        ID="TP1",
        description="Transport Process 1",
        time_model_id="tm_transport",
        type=ProcessTypeEnum.TransportProcesses,
    )


@pytest.fixture
def sample_product(sample_transport_process):
    """Create a sample product for testing."""
    return ProductData(
        ID="Product_1",
        description="Product 1",
        product_type="Product_1",
        processes={"P1": []},
        transport_process="TP1",
    )


@pytest.fixture
def sample_queue():
    """Create a sample queue for testing."""
    return QueueData(
        ID="SinkQueue",
        description="Sink Queue",
        capacity=10,
    )


@pytest.fixture
def sample_production_system(sample_time_model, sample_transport_time_model, sample_process, sample_transport_process, sample_product, sample_queue):
    """Create a minimal production system for testing."""
    return ProductionSystemData(
        time_model_data=[sample_time_model, sample_transport_time_model],
        process_data=[sample_process, sample_transport_process],
        resource_data=[],
        product_data=[sample_product],
        source_data=[],
        sink_data=[],
        state_data=[],
        port_data=[sample_queue],
        primitive_data=[],
        dependency_data=[],
        node_data=[],
        scenario_data=None,
    )


class TestSinkData:
    """Tests for SinkData."""

    def test_creation_with_required_fields(self):
        """Test creating SinkData with required fields."""
        sink = SinkData(
            ID="SK1",
            description="Sink 1",
            location=[50.0, 50.0],
            product_type="Product_1",
        )
        assert sink.ID == "SK1"
        assert sink.description == "Sink 1"
        assert sink.location == [50.0, 50.0]
        assert sink.product_type == "Product_1"

    def test_default_ports(self):
        """Test that ports defaults to empty list."""
        sink = SinkData(
            ID="SK1",
            description="Sink 1",
            location=[50.0, 50.0],
            product_type="Product_1",
        )
        assert sink.ports is None

    def test_with_ports(self):
        """Test creating SinkData with ports."""
        sink = SinkData(
            ID="SK1",
            description="Sink 1",
            location=[50.0, 50.0],
            product_type="Product_1",
            ports=["SinkQueue"],
        )
        assert sink.ports == ["SinkQueue"]

    def test_hash(self, sample_production_system, sample_product, sample_queue):
        """Test hash method."""
        sink1 = SinkData(
            ID="SK1",
            description="Sink 1",
            location=[50.0, 50.0],
            product_type="Product_1",
            ports=["SinkQueue"],
        )
        sink2 = SinkData(
            ID="SK2",
            description="Sink 2",
            location=[50.0, 50.0],
            product_type="Product_1",
            ports=["SinkQueue"],
        )
        
        hash1 = sink1.hash(sample_production_system)
        hash2 = sink2.hash(sample_production_system)
        
        # Same parameters should produce same hash
        assert hash1 == hash2

    def test_inheritance_from_locatable(self):
        """Test that SinkData inherits from Locatable."""
        sink = SinkData(
            ID="SK1",
            description="Sink 1",
            location=[50.0, 50.0],
            product_type="Product_1",
        )
        from prodsys.models.core_asset import Locatable
        assert isinstance(sink, Locatable)
        assert hasattr(sink, "location")




import pytest
from prodsys.models.production_system_data import ProductionSystemData
import prodsys.express as psx
from prodsys import runner


@pytest.fixture
def simulation_adapter() -> ProductionSystemData:
    """
    Test system with lot dependencies (batch processing) and workpiece carriers.
    
    This test validates:
    - LotDependency with carriers (batching transport with carriers)
    - LotDependency without carriers (batching production)
    - PrimitiveDependency (workpiece carriers)
    - Integration of multiple dependency types
    """
    t1 = psx.FunctionTimeModel("constant", 1.8, 0, "t1")

    p1 = psx.ProductionProcess(t1, "p1")

    t3 = psx.DistanceTimeModel(speed=50, reaction_time=0.01, ID="t3")

    carrier_tp = psx.TransportProcess(t3, "carrier_tp")

    work_piece_carrier_storage = psx.Queue(ID="work_piece_carrier_storage", location=[0, 0], capacity=2)

    work_piece_carriers = psx.Primitive(
        ID="work_piece_carriers",
        transport_process=carrier_tp,
        storages=[work_piece_carrier_storage],
        quantity_in_storages=[2],
    )

    lot_dependency_with_carrier = psx.LotDependency(
        min_lot_size=1,
        max_lot_size=4,
        ID="lot_dependency_with_carrier",
    )
    work_piece_carrier_dependency = psx.ToolDependency(
        required_entity=work_piece_carriers,
        per_lot=True
    )

    lot_dependency_without_carrier = psx.LotDependency(
        min_lot_size=2,
        max_lot_size=3,
        ID="lot_dependency_without_carrier",
    )

    machine = psx.Resource(
        [p1],
        [5, 0],
        capacity=5,
        ID="machine",
        dependencies=[lot_dependency_without_carrier],
    )

    tp = psx.TransportProcess(t3, "tp", dependencies=[lot_dependency_with_carrier, work_piece_carrier_dependency])

    transport = psx.Resource([tp], [0, 0], 4, ID="transport")
    carrier_transport = psx.Resource([carrier_tp], [0, 0], 1, ID="carrier_transport")

    product1 = psx.Product([p1], tp, "product1")

    sink1 = psx.Sink(product1, [10, 0], "sink1")

    arrival_model_1 = psx.FunctionTimeModel("exponential", 1, ID="arrival_model_1")

    source1 = psx.Source(product1, arrival_model_1, [0, 0], ID="source_1")

    system = psx.ProductionSystem([machine, transport, carrier_transport], [source1], [sink1], [work_piece_carriers])
    system.validate()
    adapter = system.to_model()
    return adapter


def test_initialize_simulation(simulation_adapter: ProductionSystemData):
    """Test that the batch controller simulation can be initialized."""
    runner_instance = runner.Runner(production_system_data=simulation_adapter)
    runner_instance.initialize_simulation()


def test_hashing(simulation_adapter: ProductionSystemData):
    """Test that the batch controller system produces a consistent hash."""
    hash_str = simulation_adapter.hash()
    assert hash_str == "bd3ba32d16ed0d31bbb099036494fb1d"


def test_run_simulation(simulation_adapter: ProductionSystemData):
    """
    Test that the batch controller simulation runs successfully with lot dependencies.
    
    This validates:
    - Products are correctly batched in lots according to lot dependencies
    - Workpiece carriers are properly managed (bound/released)
    - Both transport with carriers and production batching work together
    - System completes simulation without errors
    """
    runner_instance = runner.Runner(production_system_data=simulation_adapter)
    runner_instance.initialize_simulation()
    runner_instance.run(1000)
    assert runner_instance.env.now == 1000
    runner_instance.print_results()
    
    post_processor = runner_instance.get_post_processor()
    
    # Check throughput - should be around 1000 products
    for kpi in post_processor.throughput_and_output_KPIs:
        if kpi.name == "output":
            assert kpi.product_type == "product1"
            assert kpi.value > 950 and kpi.value < 1050
    
    # Check machine utilization - with batching (lot size 2-3), utilization is lower
    for kpi in post_processor.machine_state_KPIS:
        if kpi.name == "productive_time" and kpi.resource == "machine":
            assert kpi.value < 75 and kpi.value > 65
        
        # Transport should have moderate utilization
        if kpi.name == "productive_time" and kpi.resource == "transport":
            assert kpi.value < 35 and kpi.value > 25
        
        # Carrier transport should have moderate utilization
        if kpi.name == "productive_time" and kpi.resource == "carrier_transport":
            assert kpi.value < 45 and kpi.value > 35
    
    # Check WIP - batching can increase WIP slightly
    for kpi in post_processor.WIP_KPIs:
        if kpi.name == "WIP" and kpi.product_type == "product1":
            assert kpi.value < 4.5 and kpi.value > 3.0
    
    # Check primitive WIP for workpiece carriers
    for kpi in post_processor.primitive_WIP_KPIs:
        if kpi.name == "primitive_WIP" and kpi.product_type == "work_piece_carriers":
            assert kpi.value < 2.0 and kpi.value > 1
    
    # Check throughput time - should be reasonable with batching
    for kpi in post_processor.aggregated_throughput_time_KPIs:
        if kpi.name == "throughput_time":
            assert kpi.value < 4.50



"""
Production flow analytics module.

This module handles production flow ratio calculations.
"""

from __future__ import annotations

from functools import cached_property

import pandas as pd

from prodsys.analytics.base import AnalyticsContext
from prodsys.analytics.data_preparation import DataPreparation
from prodsys.analytics.throughput import ThroughputAnalytics

import logging

logger = logging.getLogger(__name__)


class ProductionFlowAnalytics:
    """
    Handles production flow ratio calculations.
    """
    
    def __init__(self, context: AnalyticsContext, data_prep: DataPreparation, throughput_analytics: ThroughputAnalytics):
        """
        Initialize production flow analytics.
        
        Args:
            context: Analytics context containing raw data and configuration.
            data_prep: Data preparation instance for accessing prepared data.
            throughput_analytics: Throughput analytics instance.
        """
        self.context = context
        self.data_prep = data_prep
        self.throughput_analytics = throughput_analytics
    
    @cached_property
    def df_production_flow_ratio(self) -> pd.DataFrame:
        """
        Calculates the production flow ratio for each product type.
        
        Returns:
            pd.DataFrame: DataFrame containing the production flow ratio for each product type.
                The DataFrame has the following columns:
                - Product_type: The type of the product.
                - Production: The percentage of time spent in production activities.
                - Transport: The percentage of time spent in transport activities.
                - Idle: The percentage of idle time.
        """
        df_finished_product = self.data_prep.df_finished_product.copy()
        
        if self.context.warm_up_cutoff:
            df_finished_product = df_finished_product.loc[
                df_finished_product["Time"] >= self.throughput_analytics.warm_up_cutoff_time
            ]
        
        # Production
        filtered_df = df_finished_product[
            df_finished_product["State Type"] == "Production"
        ]
        df_production = filtered_df[
            ["Product", "Product_type", "State Type", "Activity", "Time"]
        ]
        grouped_production_df = (
            df_production.groupby(["Product", "Product_type", "Activity"])["Time"]
            .sum()
            .reset_index()
        )
        pivot_production_df = grouped_production_df.pivot(
            index=["Product", "Product_type"], columns="Activity", values="Time"
        )
        pivot_production_df = pivot_production_df.fillna(0)
        pivot_production_df["Production Time"] = (
            pivot_production_df["end state"]
            + pivot_production_df.get("end interrupt", 0)
            - pivot_production_df["start state"]
            - (
                pivot_production_df.get("end interrupt", 0)
                - pivot_production_df.get("start interrupt", 0)
            )
        )
        mean_production_time = (
            pivot_production_df.groupby("Product_type")["Production Time"]
            .mean()
            .reset_index()
        )
        
        # Transport
        df_transport = df_finished_product[
            df_finished_product["State Type"] == "Transport"
        ]
        df_transport = df_transport[
            ["Product", "Product_type", "State Type", "Activity", "Time"]
        ]
        df_transport = (
            df_transport.groupby(["Product", "Product_type", "Activity"])["Time"]
            .sum()
            .reset_index()
        )
        df_transport = df_transport.pivot(
            index=["Product", "Product_type"], columns="Activity", values="Time"
        )
        df_transport = df_transport.fillna(0)
        df_transport["Transport Time"] = (
            df_transport["end state"]
            + df_transport.get("end interrupt", 0)
            - df_transport["start state"]
            - (
                df_transport.get("end interrupt", 0)
                - df_transport.get("start interrupt", 0)
            )
        )
        mean_transport_time = (
            df_transport.groupby("Product_type")["Transport Time"].mean().reset_index()
        )
        
        df_aggregated_throughput_time_copy = self.throughput_analytics.df_aggregated_throughput_time.copy()
        
        merged_df = pd.merge(
            df_aggregated_throughput_time_copy,
            mean_production_time,
            on="Product_type",
            how="inner",
        )
        merged_df = pd.merge(
            merged_df, mean_transport_time, on="Product_type", how="inner"
        )
        merged_df["Idle Time"] = (
            merged_df["Throughput_time"]
            - merged_df["Production Time"]
            - merged_df["Transport Time"]
        )
        
        percentage_df = pd.DataFrame(
            {
                "Product_type": merged_df["Product_type"],
                "Production ": merged_df["Production Time"]
                / merged_df["Throughput_time"]
                * 100,
                "Transport ": merged_df["Transport Time"]
                / merged_df["Throughput_time"]
                * 100,
                "Idle ": merged_df["Idle Time"] / merged_df["Throughput_time"] * 100,
            }
        )
        
        return percentage_df



import prodsys
import os
from prodsys.models.production_system_data import ProductionSystemData, add_default_queues_to_production_system
from prodsys.util.node_link_generation import node_link_generation


def main():

    # Time models
    ftmp1 = prodsys.time_model_data.FunctionTimeModelData(
        ID="ftmp1",
        description="function time model process 1",
        distribution_function=prodsys.time_model_data.FunctionTimeModelEnum.Normal,
        location=0.833,
        scale=0.0833
    )
    ftmp2 = prodsys.time_model_data.FunctionTimeModelData(
        ID="ftmp2",
        description="function time model process 2",
        distribution_function=prodsys.time_model_data.FunctionTimeModelEnum.Normal,
        location=4.167,
        scale=0.4167
    )
    ftm1 = prodsys.time_model_data.FunctionTimeModelData(
        ID="ftm1",
        description="function time model product 1",
        distribution_function=prodsys.time_model_data.FunctionTimeModelEnum.Exponential,
        location=10.05,
        scale=0.0
    )
    md1 = prodsys.time_model_data.DistanceTimeModelData(
        ID="md1",
        description="manhattan time model 1",
        speed=60.0,
        reaction_time=0.033,
        metric="manhattan",
    )

    # Processes
    P1 = prodsys.processes_data.ProductionProcessData(
        ID="P1",
        description="Process 1",
        time_model_id="ftmp1",
        type=prodsys.processes_data.ProcessTypeEnum.ProductionProcesses,
    )
    P2 = prodsys.processes_data.ProductionProcessData(
        ID="P2",
        description="Process 2",
        time_model_id="ftmp2",
        type=prodsys.processes_data.ProcessTypeEnum.ProductionProcesses,
    )
    TP1 = prodsys.processes_data.LinkTransportProcessData( #empty links, to be generated
        ID="TP1",
        description="Transport Process 1",
        time_model_id="md1",
        capability="TP1",
        type=prodsys.processes_data.ProcessTypeEnum.LinkTransportProcesses,
        links=[],
        can_move=True
    )

    # Resource
    R1 = prodsys.resource_data.ResourceData( #resource with two ports
        ID="R1",
        description="Resource 1",
        capacity=2,
        location=[-100, -100],
        controller=prodsys.resource_data.ControllerEnum.PipelineController,
        control_policy=prodsys.resource_data.ResourceControlPolicy.FIFO,
        process_ids=["P1", "P2"],
        #ports=[Port0.ID, Port1.ID], 
    )
    R2 = prodsys.resource_data.ResourceData( #resource without ports: standart port are generated below
        ID="R2",
        description="Resource 1_1",
        capacity=2,
        location=[0, -50.0],
        controller=prodsys.resource_data.ControllerEnum.PipelineController,
        control_policy=prodsys.resource_data.ResourceControlPolicy.FIFO,
        process_ids=["P1", "P2"],
    )
    #Transport Resource like AGV
    TR1 = prodsys.resource_data.ResourceData(
        ID="TR1",
        description="Transport Resource 1",
        capacity=1,
        location=[0, 50],
        controller=prodsys.resource_data.ControllerEnum.PipelineController,
        control_policy=prodsys.resource_data.TransportControlPolicy.SPT_transport,
        process_ids=["TP1"],
        can_move=True,
    )

    # Product
    Product_1 = prodsys.product_data.ProductData(
        ID="Product_1",
        description="",
        type="Product_1",
        processes=["P1"],
        transport_process="TP1",
    )

    # Source
    S1 = prodsys.source_data.SourceData(
        ID="S1",
        description="Source 1",
        location=[-100.0, -100.0],
        product_type="Product_1",
        time_model_id="ftm1",
        routing_heuristic=prodsys.source_data.RoutingHeuristic.shortest_queue
    )

    # Sink
    K1 = prodsys.sink_data.SinkData(
        ID="K1",
        description="Sink 1",
        location=[100.0, 100.0],
        product_type="Product_1",
    )

    
    # Assemble the production system
    production_system_instance = ProductionSystemData(
        time_model_data=[ftmp1, ftmp2, ftm1, md1],
        process_data=[P1, P2, TP1],
        #port_data=[Port0, Port1],
        resource_data=[R1, TR1, R2],
        product_data=[Product_1],
        source_data=[S1],
        sink_data=[K1],
        valid_configuration=True,
        node_data=[]
    )
    add_default_queues_to_production_system(production_system_instance, reset=False)

    BASE_DIR = os.path.dirname(os.path.abspath(__file__))
    path = os.path.join(BASE_DIR, "simulation_example_data", "FreeSpace.drawio.xml")
    node_link_generation.generate_and_apply_network(production_system_instance, path, visualize=True, style="random", simple_connection=False)
    prodsys.set_logging("DEBUG")
    runner = prodsys.runner.Runner(production_system_data=production_system_instance)
    runner.initialize_simulation()
    runner.run(500)
    runner.print_results()
    #runner.plot_results()
    #runner.save_results_as_csv() 

if __name__ == "__main__":
    main()



from enum import Enum
from typing import Union, Optional
from hashlib import md5

from pydantic import ConfigDict, Field

from prodsys.models.core_asset import CoreAsset, Locatable, Location2D

class PortInterfaceType(str, Enum):
    """
    Enum that represents the type of an interface.
    """
    INPUT = "input"
    OUTPUT = "output"
    INPUT_OUTPUT = "input_output"

class PortType(str, Enum):
    """
    Enum that represents the kind of an interface.
    """
    QUEUE = "queue"
    STORE = "store"

class QueueData(CoreAsset):
    """
    Class that represents a queue for products at a resource / sink / source. If capacity is 0, the queue is considered infinite. Otherwise, the queue can hold a finite number of products cooresponding to the capacity.

    Args:
        ID (str): ID of the queue.
        description (str): Description of the queue.
        capacity (Union[int, float]): Capacity of the queue. If 0, the queue is considered infinite. Otherwise, the queue can hold a finite number of products cooresponding to the capacity.
        location (List[float]): Location of the queue. It has to be a list of length 2. If not provided, the position is inferred from the parent resource. Defaults to None.
        interface_type (PortInterfaceType): Type of the interface. Defaults to PortInterfaceType.INPUT_OUTPUT.
        port_type (PortType): Type of the port. Defaults to PortType.QUEUE.
        dependency_ids (List[str]): List of dependency IDs that are required by the queue.

    Examples:
        A finite queue with ID "Q1", description "Queue 1" and capacity 10:
        ``` py
        import prodsys
        prodsys.queue_data.QueueData(
            ID="Q1",
            description="Queue 1",
            capacity=10,
        )
        ```
        An infinite queue with ID "Q1", description "Queue 1" and capacity 0:
        ``` py
        import prodsys
        prodsys.queue_data.QueueData(
            ID="Q1",
            description="Queue 1",
            capacity=0,
        )
        ```
    """
    capacity: Union[int, float] = 0.0
    location: Optional[Location2D] = Field(
        default=None,
    )
    interface_type: PortInterfaceType = PortInterfaceType.INPUT_OUTPUT
    port_type: PortType = Field(default=PortType.QUEUE, init=False, frozen=True)
    dependency_ids: list[str] = []


    def hash(self) -> str:
        """
        Returns a unique hash for the queue considering its capacity.


        Returns:
            str: Hash of the queue.
        """
        # TODO: consider dependencies
        return md5((str(self.capacity)).encode("utf-8")).hexdigest()

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "ID": "Q1",
                    "description": "Finte Queue",
                    "capacity": 10,
                },
                {
                    "ID": "Q1",
                    "description": "Infinite Queue",
                    "capacity": 0.0,
                },
            ]
        }
    )


class StoreData(QueueData):
    """
    Class that represents a store which is a storage with multiple input and output port locations.

    If capacity is 0, the store is considered infinite. Otherwise, the queue can hold a finite number of products cooresponding to the capacity.

    Args:
        ID (str): ID of the store.
        description (str): Description of the store.
        capacity (Union[int, float]): Capacity of the store. If 0, the queue is considered infinite. Otherwise, the queue can hold a finite number of products cooresponding to the capacity.
        location (List[float]): Location of the store. It has to be a list of length 2.
        port_locations (Optional[list[list[float]]], optional): List of port locations for input and output. Each port location has to be a list of length 2. Defaults to None.

    Examples:
        A finite store with ID "ST1", description "Store Q1" and capacity 10:
        ``` py
        import prodsys
        prodsys.queue_data.StoreData(
            ID="ST1",
            description="Store Q1",
            capacity=10,
            location=[10, 0],
        )
        ```
        An infinite store with ID "ST1", description "Store Q1", capacity 0 and port locations [[11, 0], [12, 0]]:
        ``` py
        import prodsys
        prodsys.queue_data.StoreData(
            ID="ST1",
            description="Store Q1",
            capacity=0,
            location=[10, 0],
            port_locations=[[11, 0], [12, 0]],
        )
        ```
    """
    port_locations: Optional[list[Location2D]] = Field(
        default=None,
    )
    interface_type: PortInterfaceType = PortInterfaceType.INPUT_OUTPUT
    port_type: PortType = Field(default=PortType.STORE, init=False, frozen=True)

    def hash(self) -> str:
        """
        Returns a unique hash for the queue considering its capacity.


        Returns:
            str: Hash of the queue.
        """
        queue_hash = QueueData.hash(self)
        location_hash = Locatable.hash(self)
        port_locations_hash = md5(
            (str(self.port_locations) if self.port_locations else "").encode("utf-8")
        ).hexdigest()
        return md5((location_hash + queue_hash + port_locations_hash).encode("utf-8")).hexdigest()

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "ID": "Q1",
                    "description": "Finite Store",
                    "capacity": 10,
                    "location": [10, 0],
                },
                {
                    "ID": "Q1",
                    "description": "Infinite Store",
                    "capacity": 0.0,
                    "location": [10, 0],
                    "port_locations": [[11, 0], [12, 0]],
                },
            ]
        }
    )


QUEUE_DATA_UNION = Union[QueueData, StoreData]



from __future__ import annotations

from typing import List, Optional, Union, Any
from uuid import uuid1

from pydantic import Field, field_validator
from pydantic.dataclasses import dataclass

from prodsys.models import product_data

from prodsys.express import core, process
from prodsys.express.dependency import Dependency


@dataclass
class Product(core.ExpressObject):
    """
    Class that represents a product.

    Args:
        process (Union[process.ProcessModel, process.SequentialProcess, List[process.ProductionProcess]]): Process model of the product required for its production. This can be a process model (DAG), sequential process, or list of production processes (for backward compatibility).
        transport_process (process.TransportProcess): Transport process of the product.
        ID (str): ID of the product.

    Attributes:
        process (Union[process.ProcessModel, process.SequentialProcess]): Process model of the product required for its production.
        transport_process (process.TransportProcess): Transport process of the product.
        ID (Optional[str]): ID of the product.
        dependencies (Optional[List[Dependency]]): Dependencies of the product.


    Examples:
        Product with process model and transport process:
        ```py
        import prodsys.express as psx
        
        # Create individual processes
        welding_time_model = psx.FunctionTimeModel("normal", 20.0, 5.0)
        welding_process = psx.ProductionProcess(time_model=welding_time_model, ID="WELD")
        
        assembly_time_model = psx.FunctionTimeModel("normal", 15.0, 3.0)
        assembly_process = psx.ProductionProcess(time_model=assembly_time_model, ID="ASSEMBLY")
        
        # Create a process model (DAG)
        process_model = psx.ProcessModel(
            time_model=psx.FunctionTimeModel("constant", 0.0),
            process_ids=["WELD", "ASSEMBLY"],
            adjacency_matrix={"WELD": ["ASSEMBLY"], "ASSEMBLY": []},
            ID="PROCESS_MODEL_1"
        )
        
        # Create transport process
        transport_time_model = psx.ManhattenDistanceTimeModel(speed=10, reaction_time=0.3)
        transport_process = psx.TransportProcess(time_model=transport_time_model)
        
        # Create product with process model
        psx.Product(
            process=process_model,
            transport_process=transport_process
        )
        ```

    """

    process: Union[process.ProcessModel]
    transport_process: Union[
        process.TransportProcess, process.RequiredCapabilityProcess
    ]
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))
    dependencies: Optional[List[Dependency]] = Field(default_factory=list)
    becomes_consumable : bool = False

    @field_validator('process', mode='before')
    @classmethod
    def validate_process(cls, v):
        """Convert list of processes to SequentialProcess for backward compatibility."""
        if isinstance(v, list):
            if len(v) == 0:
                raise ValueError("Process list cannot be empty")
            # Extract process IDs
            process_ids = [p.ID for p in v]
            
            # Create SequentialProcess with a dummy time model
            from prodsys.express import time_model
            dummy_time_model = time_model.FunctionTimeModel("constant", 0.0, ID="dummy_time_model")
            # TODO: update this function to work properly with ProcessModel
            sequential_process = process.SequentialProcess(
                time_model=dummy_time_model,
                process_ids=process_ids,
                ID="sequential_process"
            )
            
            return sequential_process
        return v

    def to_model(self) -> product_data.ProductData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            product_data.ProductData: An instance of the data object.
        """
        return product_data.ProductData(
            ID=self.ID,
            type=self.ID,
            description="",
            processes=self.process.adjacency_matrix,
            transport_process=self.transport_process.ID,
            dependency_ids=[dependency.ID for dependency in self.dependencies],
            becomes_consumable= self.becomes_consumable,
        )


def create_product(*args, **kwargs):
    """
    Factory function to create Product instances with backward compatibility.
    
    Supports both old and new API patterns:
    - Old: create_product([processes], transport_process, ID)
    - Old: create_product(processes=[...], transport_process=..., ID=...)
    - New: create_product(process=..., transport_process=..., ID=...)
    """
    # Handle old positional API: Product([processes], transport_process, ID)
    if args and not kwargs:
        if len(args) >= 2:
            processes_arg = args[0]
            transport_process_arg = args[1]
            id_arg = args[2] if len(args) > 2 else None
            
            # Handle old API with processes parameter
            if isinstance(processes_arg, list):
                kwargs['process'] = processes_arg
                kwargs['transport_process'] = transport_process_arg
                if id_arg:
                    kwargs['ID'] = id_arg
            else:
                # Handle new API with single process
                kwargs['process'] = processes_arg
                kwargs['transport_process'] = transport_process_arg
                if id_arg:
                    kwargs['ID'] = id_arg
        else:
            raise ValueError("Product requires at least process and transport_process arguments")
    
    # Handle old API with processes keyword
    if 'processes' in kwargs and 'process' not in kwargs:
        kwargs['process'] = kwargs.pop('processes')
    
    return Product(**kwargs)





"""
Unit tests for ProcessMatcher compatibility table methods.

This module contains comprehensive tests for the modularized compatibility table
precomputation methods in ProcessMatcher.
"""

import pytest
from unittest.mock import Mock, patch, MagicMock
from typing import Dict, List, Tuple

import prodsys.express as psx
from prodsys.models.production_system_data import ProductionSystemData
from prodsys.simulation.process_matcher import (
    ProcessMatcher,
    ResourceCompatibilityKey,
    TransportCompatibilityKey,
)
from prodsys.simulation import process, request, resources
from prodsys.simulation.entities import product
from prodsys.models.source_data import RoutingHeuristic


class TestProcessMatcherCompatibility:
    """Test class for ProcessMatcher compatibility table methods."""

    @pytest.fixture
    def most_trivial_system(self) -> ProductionSystemData:
        """Create a most trivial production system for testing."""
        t1 = psx.FunctionTimeModel("constant", 0.8, 0, "t1")
        p1 = psx.ProductionProcess(t1, "p1")
        t3 = psx.FunctionTimeModel("normal", 0.1, 0.01, ID="t3")
        tp = psx.TransportProcess(t3, "tp")

        machine = psx.Resource([p1], [5, 0], 1, ID="machine")
        transport = psx.Resource([tp], [0, 0], 1, ID="transport")

        product1 = psx.Product([p1], tp, "product1")
        sink1 = psx.Sink(product1, [10, 0], "sink1")
        arrival_model_1 = psx.FunctionTimeModel("exponential", 1, ID="arrival_model_1")
        source1 = psx.Source(product1, arrival_model_1, [0, 0], ID="source_1")

        system = psx.ProductionSystem([machine, transport], [source1], [sink1])
        return system.to_model()

    @pytest.fixture
    def primitive_complex_system(self) -> ProductionSystemData:
        """Create a primitive complex production system for testing."""
        t1 = psx.FunctionTimeModel("exponential", 0.8, 0, "t1")
        t2 = psx.FunctionTimeModel("exponential", 1.2, 0, "t2")
        p1 = psx.ProductionProcess(t1, "p1")
        p2 = psx.ProductionProcess(t2, "p2")

        t3 = psx.DistanceTimeModel(60, 0.05, "manhattan", ID="t3")
        tp = psx.TransportProcess(t3, "tp")
        tp_prim = psx.TransportProcess(t3, "tp_primitive")

        machine = psx.Resource([p1], [5, 0], 1, ID="machine")
        machine2 = psx.Resource([p2], [5, 5], 2, ID="machine2")
        transport = psx.Resource([tp], [3, 0], 1, ID="transport")
        transport2 = psx.Resource([tp], [3, 0], 1, ID="transport2")
        transport_prim = psx.Resource([tp_prim], [4, 0], 1, ID="transport_primitive")

        storage1 = psx.Store(ID="storage1", location=[6, 0], capacity=30)
        storage2 = psx.Store(ID="storage2", location=[11, 0], capacity=20)

        primitive1 = psx.Primitive(
            ID="primitive1",
            transport_process=tp_prim,
            storages=[storage1],
            quantity_in_storages=[10],
        )

        primitive2 = psx.Primitive(
            ID="primitive2",
            transport_process=tp_prim,
            storages=[storage2],
            quantity_in_storages=[20],
        )

        primitive_dependency_1 = psx.ToolDependency(
            ID="primitive_dependency_1",
            required_entity=primitive1,
        )
        primitive_dependency_2 = psx.ToolDependency(
            ID="primitive_dependency_2",
            required_entity=primitive2,
        )

        product1 = psx.Product(
            process=[p1, p2],
            transport_process=tp,
            ID="product1",
            dependencies=[primitive_dependency_1],
        )
        product2 = psx.Product(
            process=[p2, p1],
            transport_process=tp,
            ID="product2",
            dependencies=[primitive_dependency_2],
        )

        sink1 = psx.Sink(product1, [10, 0], "sink1")
        sink2 = psx.Sink(product2, [10, 0], "sink2")

        arrival_model_1 = psx.FunctionTimeModel("constant", 0.9, ID="arrival_model_1")
        arrival_model_2 = psx.FunctionTimeModel("constant", 1.1, ID="arrival_model_2")

        source1 = psx.Source(product1, arrival_model_1, [0, 0], ID="source_1")
        source2 = psx.Source(product2, arrival_model_2, [0, 0], ID="source_2")

        system = psx.ProductionSystem(
            [machine, machine2, transport, transport2, transport_prim],
            [source1, source2],
            [sink1, sink2],
            [primitive1, primitive2],
        )
        return system.to_model()

    @pytest.fixture
    def link_transport_with_capabilities_system(self) -> ProductionSystemData:
        """Create a link transport system with capabilities for testing."""
        time_model_agv = psx.DistanceTimeModel(
            speed=90, reaction_time=0.2, ID="time_model_x"
        )
        time_model_machine1 = psx.FunctionTimeModel(
            distribution_function="constant", location=3, ID="time_model_ap23"
        )

        node1 = psx.Node(location=[0, 20], ID="node1")
        node2 = psx.Node(location=[50, 20], ID="node2")
        node3 = psx.Node(location=[100, 20], ID="node3")

        rcp01 = psx.RequiredCapabilityProcess(
            time_model=time_model_agv, capability="euro_palette_transport", ID="rtp01"
        )
        productionprocess01 = psx.ProductionProcess(
            time_model=time_model_machine1, ID="pp01"
        )

        machine01 = psx.Resource(
            ID="resource01",
            processes=[productionprocess01],
            location=[0, 0],
        )

        product01 = psx.Product(
            process=[productionprocess01],
            transport_process=rcp01,
            ID="product01",
        )

        source01 = psx.Source(
            product=product01,
            ID="source01",
            time_model=psx.FunctionTimeModel("constant", 6, ID="tm_source01"),
            location=[-10, 0],
        )
        sink01 = psx.Sink(product=product01, ID="sink01", location=[-10, 100])

        ltp01_links = [
            [source01, machine01],
            [machine01, node1],
            [node1, node2],
            [node2, node3],
        ]

        ltp01 = psx.LinkTransportProcess(
            time_model=time_model_agv,
            capability="euro_palette_transport",
            ID="ltp01",
            links=ltp01_links,
        )

        agv01 = psx.Resource(
            location=[50, 20],
            ID="agv01",
            processes=[ltp01],
        )

        productionsystem = psx.ProductionSystem(
            resources=[agv01, machine01],
            sources=[source01],
            sinks=[sink01],
            ID="productionsystem01",
        )
        return productionsystem.to_model()

    @pytest.fixture
    def link_transport_without_capabilities_system(self) -> ProductionSystemData:
        """Create a link transport system without capabilities for testing."""
        time_model_agv = psx.DistanceTimeModel(
            speed=360, reaction_time=0, ID="time_model_x"
        )
        time_model_machine1 = psx.FunctionTimeModel(
            distribution_function="constant", location=3, ID="time_model_ap23"
        )

        node1 = psx.Node(location=[10, 0], ID="node1")
        node2 = psx.Node(location=[0, 15], ID="node2")

        ltp01 = psx.LinkTransportProcess(time_model=time_model_agv, ID="ltp01")
        productionprocess01 = psx.ProductionProcess(
            time_model=time_model_machine1, ID="pp01"
        )

        machine01 = psx.Resource(
            ID="resource01",
            processes=[productionprocess01],
            location=[10, 10],
        )

        agv01 = psx.Resource(
            location=[0, 0],
            ID="agv01",
            processes=[ltp01],
        )

        product01 = psx.Product(
            process=[productionprocess01],
            transport_process=ltp01,
            ID="product01",
        )

        source01 = psx.Source(
            product=product01,
            ID="source01",
            time_model=psx.FunctionTimeModel("constant", 6, ID="tm_source01"),
            location=[0, 0],
        )
        sink01 = psx.Sink(product=product01, ID="sink01", location=[20, 25])

        # Update processes
        links = [
            [source01, node1],
            [node1, machine01],
            [machine01, node2],
            [node2, sink01],
        ]
        ltp01.set_links(links)

        productionsystem = psx.ProductionSystem(
            resources=[agv01, machine01],
            sources=[source01],
            sinks=[sink01],
            ID="productionsystem01",
        )
        return productionsystem.to_model()

    @pytest.fixture
    def process_matcher_most_trivial(self, most_trivial_system):
        """Create a ProcessMatcher instance for most trivial system."""
        from prodsys import runner

        runner_instance = runner.Runner(production_system_data=most_trivial_system)
        runner_instance.initialize_simulation()
        return runner_instance.product_factory.router.request_handler.process_matcher

    @pytest.fixture
    def process_matcher_primitive_complex(self, primitive_complex_system):
        """Create a ProcessMatcher instance for primitive complex system."""
        from prodsys import runner

        runner_instance = runner.Runner(production_system_data=primitive_complex_system)
        runner_instance.initialize_simulation()
        return runner_instance.product_factory.router.request_handler.process_matcher

    @pytest.fixture
    def process_matcher_link_transport_with_capabilities(
        self, link_transport_with_capabilities_system
    ):
        """Create a ProcessMatcher instance for link transport with capabilities system."""
        from prodsys import runner

        runner_instance = runner.Runner(
            production_system_data=link_transport_with_capabilities_system
        )
        runner_instance.initialize_simulation()
        return runner_instance.product_factory.router.request_handler.process_matcher

    @pytest.fixture
    def process_matcher_link_transport_without_capabilities(
        self, link_transport_without_capabilities_system
    ):
        """Create a ProcessMatcher instance for link transport without capabilities system."""
        from prodsys import runner

        runner_instance = runner.Runner(
            production_system_data=link_transport_without_capabilities_system
        )
        runner_instance.initialize_simulation()
        return runner_instance.product_factory.router.request_handler.process_matcher

    def test_create_dummy_products(self, process_matcher_most_trivial):
        """Test _create_dummy_products method."""
        dummy_products = process_matcher_most_trivial._create_dummy_products()

        assert isinstance(dummy_products, dict)
        assert len(dummy_products) == 1  # One product type in most trivial system
        assert "product1" in dummy_products
        assert isinstance(dummy_products["product1"], product.Product)

    def test_precompute_production_compatibility_most_trivial(
        self, process_matcher_most_trivial
    ):
        """Test production compatibility precomputation for most trivial system."""
        # Clear existing compatibility tables
        process_matcher_most_trivial.production_compatibility = {}

        dummy_products = process_matcher_most_trivial._create_dummy_products()
        process_matcher_most_trivial._precompute_production_compatibility(
            dummy_products
        )

        # Check that production compatibility table is populated
        assert len(process_matcher_most_trivial.production_compatibility) > 0

        # Check for specific process signatures
        production_keys = list(
            process_matcher_most_trivial.production_compatibility.keys()
        )
        assert any("ProductionProcess:p1" in str(key) for key in production_keys)

    def test_precompute_production_compatibility_primitive_complex(
        self, process_matcher_primitive_complex
    ):
        """Test production compatibility precomputation for primitive complex system."""
        # Clear existing compatibility tables
        process_matcher_primitive_complex.production_compatibility = {}

        dummy_products = process_matcher_primitive_complex._create_dummy_products()
        process_matcher_primitive_complex._precompute_production_compatibility(
            dummy_products
        )

        # Check that production compatibility table is populated
        assert len(process_matcher_primitive_complex.production_compatibility) > 0

        # Check for specific process signatures
        production_keys = list(
            process_matcher_primitive_complex.production_compatibility.keys()
        )
        assert any("ProductionProcess:p1" in str(key) for key in production_keys)
        assert any("ProductionProcess:p2" in str(key) for key in production_keys)

    def test_precompute_transport_compatibility_most_trivial(
        self, process_matcher_most_trivial
    ):
        """Test transport compatibility precomputation for most trivial system."""
        # Clear existing compatibility tables
        process_matcher_most_trivial.transport_compatibility = {}
        process_matcher_most_trivial.reachability_cache = {}
        process_matcher_most_trivial.route_cache = {}

        dummy_products = process_matcher_most_trivial._create_dummy_products()
        process_matcher_most_trivial._precompute_transport_compatibility(dummy_products)

        # Check that transport compatibility table is populated
        assert len(process_matcher_most_trivial.transport_compatibility) > 0
        assert len(process_matcher_most_trivial.reachability_cache) > 0

    def test_precompute_transport_compatibility_link_transport_with_capabilities(
        self, process_matcher_link_transport_with_capabilities
    ):
        """Test transport compatibility precomputation for link transport with capabilities system."""
        # Clear existing compatibility tables
        process_matcher_link_transport_with_capabilities.transport_compatibility = {}
        process_matcher_link_transport_with_capabilities.reachability_cache = {}
        process_matcher_link_transport_with_capabilities.route_cache = {}

        dummy_products = (
            process_matcher_link_transport_with_capabilities._create_dummy_products()
        )
        process_matcher_link_transport_with_capabilities._precompute_transport_compatibility(
            dummy_products
        )

        # Check that transport compatibility table is populated
        assert (
            len(
                process_matcher_link_transport_with_capabilities.transport_compatibility
            )
            > 0
        )
        assert (
            len(process_matcher_link_transport_with_capabilities.reachability_cache) > 0
        )

        # Verify route cache has entries
        assert len(process_matcher_link_transport_with_capabilities.route_cache) > 0, (
            "Route cache should contain route entries"
        )

        # Verify that nodes are included in reachability cache
        all_locations = (
            process_matcher_link_transport_with_capabilities._get_all_locations()
        )
        location_ids = [loc.data.ID for loc in all_locations]

        # Check that nodes are included in locations
        assert "node1" in location_ids, "node1 should be included in all_locations"
        assert "node2" in location_ids, "node2 should be included in all_locations"
        assert "node3" in location_ids, "node3 should be included in all_locations"

        expected_node_connections = [
            ("source01_default_output_queue", "source01_default_output_queue"),
            ("source01_default_output_queue", "resource01_default_queue"),
            ("resource01_default_queue", "source01_default_output_queue"),
            ("resource01_default_queue", "resource01_default_queue"),
        ]
        for origin_id, target_id in expected_node_connections:
            cache_key = (origin_id, target_id)
            assert (
                cache_key
                in process_matcher_link_transport_with_capabilities.reachability_cache
            ), f"Missing reachability cache entry for {origin_id} -> {target_id}"
            assert (
                process_matcher_link_transport_with_capabilities.reachability_cache[
                    cache_key
                ]
                is True
            ), f"Reachability cache entry for {origin_id} -> {target_id} should be True"

    def test_precompute_transport_compatibility_link_transport_without_capabilities(
        self, process_matcher_link_transport_without_capabilities
    ):
        """Test transport compatibility precomputation for link transport without capabilities system."""
        # Clear existing compatibility tables
        process_matcher_link_transport_without_capabilities.transport_compatibility = {}
        process_matcher_link_transport_without_capabilities.reachability_cache = {}
        process_matcher_link_transport_without_capabilities.route_cache = {}

        dummy_products = (
            process_matcher_link_transport_without_capabilities._create_dummy_products()
        )
        process_matcher_link_transport_without_capabilities._precompute_transport_compatibility(
            dummy_products
        )

        # Check that transport compatibility table is populated
        assert (
            len(
                process_matcher_link_transport_without_capabilities.transport_compatibility
            )
            > 0
        )
        assert (
            len(process_matcher_link_transport_without_capabilities.reachability_cache)
            > 0
        )

    def test_precompute_rework_compatibility(self, process_matcher_most_trivial):
        """Test rework compatibility precomputation."""
        # Clear existing compatibility tables
        process_matcher_most_trivial.rework_compatibility = {}

        process_matcher_most_trivial._precompute_rework_compatibility()

        # Check that rework compatibility table is populated (may be empty if no rework processes)
        assert isinstance(process_matcher_most_trivial.rework_compatibility, dict)

    def test_get_all_locations(self, process_matcher_most_trivial):
        """Test _get_all_locations method."""
        locations = process_matcher_most_trivial._get_all_locations()

        assert isinstance(locations, list)
        assert len(locations) > 0

        # Check that we have resources, sources, and sinks
        location_ids = [loc.data.ID for loc in locations]
        assert "machine" in location_ids
        assert "source_1" in location_ids
        assert "sink1" in location_ids

    def test_get_all_queues(self, process_matcher_most_trivial):
        """Test _get_all_queues method."""
        queues = process_matcher_most_trivial._get_all_queues()

        assert isinstance(queues, list)
        # In most trivial system, there should be queues for resources

    def test_add_to_transport_compatibility(self, process_matcher_most_trivial):
        """Test _add_to_transport_compatibility method."""
        # Clear existing compatibility tables
        process_matcher_most_trivial.transport_compatibility = {}

        # Create a mock transport resource and process
        mock_resource = Mock(spec=resources.Resource)
        mock_process = Mock(spec=process.TransportProcess)

        key = TransportCompatibilityKey(
            origin_id="source_1",
            target_id="sink1",
            process_signature="TransportProcess:tp",
        )

        process_matcher_most_trivial._add_to_transport_compatibility(
            key, mock_resource, mock_process
        )

        # Check that the entry was added
        assert key in process_matcher_most_trivial.transport_compatibility
        assert (
            mock_resource,
            mock_process,
        ) in process_matcher_most_trivial.transport_compatibility[key]

    def test_add_to_transport_compatibility_duplicate_prevention(
        self, process_matcher_most_trivial
    ):
        """Test that _add_to_transport_compatibility prevents duplicates."""
        # Clear existing compatibility tables
        process_matcher_most_trivial.transport_compatibility = {}

        # Create a mock transport resource and process
        mock_resource = Mock(spec=resources.Resource)
        mock_process = Mock(spec=process.TransportProcess)

        key = TransportCompatibilityKey(
            origin_id="source_1",
            target_id="sink1",
            process_signature="TransportProcess:tp",
        )

        # Add the same entry twice
        process_matcher_most_trivial._add_to_transport_compatibility(
            key, mock_resource, mock_process
        )
        process_matcher_most_trivial._add_to_transport_compatibility(
            key, mock_resource, mock_process
        )

        # Check that only one entry exists
        assert len(process_matcher_most_trivial.transport_compatibility[key]) == 1

    def test_handle_required_capability_process(
        self, process_matcher_link_transport_with_capabilities
    ):
        """Test _handle_required_capability_process method."""
        # Create mock objects
        mock_request = Mock(spec=request.Request)
        mock_requested_process = Mock(spec=process.RequiredCapabilityProcess)
        mock_requested_process.data = Mock()
        mock_requested_process.data.capability = "euro_palette_transport"
        mock_requested_process.get_process_signature.return_value = (
            "RequiredCapabilityProcess:rtp01"
        )

        mock_resource = Mock(spec=resources.Resource)
        mock_offered_process = Mock(spec=process.RequiredCapabilityProcess)
        mock_offered_process.data = Mock()
        mock_offered_process.data.capability = "euro_palette_transport"

        mock_origin = Mock()
        mock_origin.data = Mock()
        mock_origin.data.ID = "source01"
        mock_target = Mock()
        mock_target.data = Mock()
        mock_target.data.ID = "sink01"

        # Clear existing compatibility tables
        process_matcher_link_transport_with_capabilities.transport_compatibility = {}
        process_matcher_link_transport_with_capabilities.reachability_cache = {}

        process_matcher_link_transport_with_capabilities._handle_required_capability_process(
            mock_request,
            mock_requested_process,
            mock_resource,
            mock_offered_process,
            mock_origin,
            mock_target,
        )

        # Check that compatibility was added
        assert (
            len(
                process_matcher_link_transport_with_capabilities.transport_compatibility
            )
            > 0
        )
        assert (
            "source01",
            "sink01",
        ) in process_matcher_link_transport_with_capabilities.reachability_cache

    def test_handle_required_capability_process_capability_mismatch(
        self, process_matcher_link_transport_with_capabilities
    ):
        """Test _handle_required_capability_process with capability mismatch."""
        # Create mock objects with different capabilities
        mock_request = Mock(spec=request.Request)
        mock_requested_process = Mock(spec=process.RequiredCapabilityProcess)
        mock_requested_process.data = Mock()
        mock_requested_process.data.capability = "euro_palette_transport"
        mock_requested_process.get_process_signature.return_value = (
            "RequiredCapabilityProcess:rtp01"
        )

        mock_resource = Mock(spec=resources.Resource)
        mock_offered_process = Mock(spec=process.RequiredCapabilityProcess)
        mock_offered_process.data = Mock()
        mock_offered_process.data.capability = "different_capability"

        mock_origin = Mock()
        mock_origin.data = Mock()
        mock_origin.data.ID = "source01"
        mock_target = Mock()
        mock_target.data = Mock()
        mock_target.data.ID = "sink01"

        # Clear existing compatibility tables
        process_matcher_link_transport_with_capabilities.transport_compatibility = {}
        process_matcher_link_transport_with_capabilities.reachability_cache = {}

        process_matcher_link_transport_with_capabilities._handle_required_capability_process(
            mock_request,
            mock_requested_process,
            mock_resource,
            mock_offered_process,
            mock_origin,
            mock_target,
        )

        # Check that no compatibility was added
        assert (
            len(
                process_matcher_link_transport_with_capabilities.transport_compatibility
            )
            == 0
        )
        assert (
            "source01",
            "sink01",
        ) not in process_matcher_link_transport_with_capabilities.reachability_cache

    @patch("prodsys.simulation.route_finder.find_route")
    def test_handle_link_transport_process(
        self, mock_find_route, process_matcher_link_transport_without_capabilities
    ):
        """Test _handle_link_transport_process method."""
        # Mock the route finder to return a valid route
        mock_route = [Mock(), Mock()]
        mock_find_route.return_value = mock_route

        # Create mock objects
        mock_request = Mock(spec=request.Request)
        mock_request.requesting_item = Mock()
        mock_requested_process = Mock(spec=process.LinkTransportProcess)
        mock_requested_process.get_process_signature.return_value = (
            "LinkTransportProcess:ltp01"
        )

        mock_resource = Mock(spec=resources.Resource)
        mock_origin = Mock()
        mock_origin.data = Mock()
        mock_origin.data.ID = "source01"
        mock_target = Mock()
        mock_target.data = Mock()
        mock_target.data.ID = "sink01"
        
        mock_offered_process = Mock(spec=process.LinkTransportProcess)
        mock_offered_process.get_process_signature.return_value = "LinkTransportProcess:ConveyorProcess_1"
        mock_offered_process.links = [[mock_origin, mock_target]]  # Set up links for the mock

        # Clear existing compatibility tables
        process_matcher_link_transport_without_capabilities.transport_compatibility = {}
        process_matcher_link_transport_without_capabilities.reachability_cache = {}
        process_matcher_link_transport_without_capabilities.route_cache = {}

        process_matcher_link_transport_without_capabilities._handle_link_transport_process(
            mock_request,
            mock_requested_process,
            mock_resource,
            mock_offered_process,
            mock_origin,
            mock_target,
        )

        # Check that compatibility was added
        assert (
            len(
                process_matcher_link_transport_without_capabilities.transport_compatibility
            )
            > 0
        )
        assert (
            "source01",
            "sink01",
        ) in process_matcher_link_transport_without_capabilities.reachability_cache
        assert len(process_matcher_link_transport_without_capabilities.route_cache) > 0

    @patch("prodsys.simulation.route_finder.find_route")
    def test_handle_link_transport_process_no_route(
        self, mock_find_route, process_matcher_link_transport_without_capabilities
    ):
        """Test _handle_link_transport_process method when no route is found."""
        # Mock the route finder to return None (no route found)
        mock_find_route.return_value = None

        # Create mock objects
        mock_request = Mock(spec=request.Request)
        mock_request.requesting_item = Mock()
        mock_requested_process = Mock(spec=process.LinkTransportProcess)
        mock_requested_process.get_process_signature.return_value = (
            "LinkTransportProcess:ltp01"
        )

        mock_resource = Mock(spec=resources.Resource)
        mock_origin = Mock()
        mock_origin.data = Mock()
        mock_origin.data.ID = "source01"
        mock_target = Mock()
        mock_target.data = Mock()
        mock_target.data.ID = "sink01"
        
        mock_offered_process = Mock(spec=process.LinkTransportProcess)
        mock_offered_process.get_process_signature.return_value = "LinkTransportProcess:ConveyorProcess_1"
        mock_offered_process.links = [[mock_origin, mock_target]]  # Set up links for the mock

        # Clear existing compatibility tables
        process_matcher_link_transport_without_capabilities.transport_compatibility = {}
        process_matcher_link_transport_without_capabilities.reachability_cache = {}
        process_matcher_link_transport_without_capabilities.route_cache = {}

        process_matcher_link_transport_without_capabilities._handle_link_transport_process(
            mock_request,
            mock_requested_process,
            mock_resource,
            mock_offered_process,
            mock_origin,
            mock_target,
        )

        # Check that no compatibility was added
        assert (
            len(
                process_matcher_link_transport_without_capabilities.transport_compatibility
            )
            == 0
        )
        assert (
            "source01",
            "sink01",
        ) not in process_matcher_link_transport_without_capabilities.reachability_cache

    def test_cache_route(self, process_matcher_most_trivial):
        """Test _cache_route method."""
        # Clear existing route cache
        process_matcher_most_trivial.route_cache = {}

        # Create mock objects
        mock_request = Mock(spec=request.Request)
        mock_origin = Mock()
        mock_origin.data = Mock()
        mock_origin.data.ID = "source_1"
        mock_target = Mock()
        mock_target.data = Mock()
        mock_target.data.ID = "sink1"
        mock_process = Mock(spec=process.TransportProcess)
        mock_process.get_process_signature.return_value = "TransportProcess:tp"

        mock_route = [Mock(), Mock()]

        process_matcher_most_trivial._cache_route(
            mock_request, mock_origin, mock_target, mock_process, mock_route
        )

        # Check that route was cached
        route_key = ("source_1", "sink1", "TransportProcess:tp")
        assert route_key in process_matcher_most_trivial.route_cache
        assert process_matcher_most_trivial.route_cache[route_key] == mock_request

    def test_precompute_compatibility_tables_integration(
        self, process_matcher_most_trivial
    ):
        """Test the complete precompute_compatibility_tables method integration."""
        # Clear all compatibility tables
        process_matcher_most_trivial.production_compatibility = {}
        process_matcher_most_trivial.transport_compatibility = {}
        process_matcher_most_trivial.reachability_cache = {}
        process_matcher_most_trivial.rework_compatibility = {}
        process_matcher_most_trivial.route_cache = {}

        # Run the complete precomputation
        process_matcher_most_trivial.precompute_compatibility_tables()

        # Check that all tables are populated
        assert len(process_matcher_most_trivial.production_compatibility) > 0
        assert len(process_matcher_most_trivial.transport_compatibility) > 0
        assert len(process_matcher_most_trivial.reachability_cache) > 0
        assert isinstance(process_matcher_most_trivial.rework_compatibility, dict)

    def test_get_compatible_production_processes(self, process_matcher_most_trivial):
        """Test get_compatible method for production processes."""
        # Get a product from the system
        dummy_products = process_matcher_most_trivial._create_dummy_products()
        product = list(dummy_products.values())[0]

        # Get all required processes
        required_processes = process_matcher_most_trivial.get_all_required_processes(
            product
        )

        # Get compatible resources
        compatible_resources = process_matcher_most_trivial.get_compatible(
            required_processes
        )

        assert isinstance(compatible_resources, list)
        assert len(compatible_resources) > 0

        # Check that all returned items are tuples of (resource, process)
        for resource, process_obj in compatible_resources:
            assert isinstance(resource, resources.Resource)
            assert isinstance(
                process_obj,
                (
                    process.ProductionProcess,
                    process.TransportProcess,
                    process.CapabilityProcess,
                    process.ReworkProcess,
                    process.CompoundProcess,
                ),
            )

    def test_get_transport_compatible(self, process_matcher_most_trivial):
        """Test get_transport_compatible method."""
        # Get locations from the system
        locations = process_matcher_most_trivial._get_all_locations()
        origin = locations[0]
        target = locations[1]

        # Get transport compatible resources
        compatible_resources = process_matcher_most_trivial.get_transport_compatible(
            origin, target, "TransportProcess:tp"
        )

        assert isinstance(compatible_resources, list)

    def test_get_rework_compatible_no_rework_processes(
        self, process_matcher_most_trivial
    ):
        """Test get_rework_compatible method when no rework processes exist."""
        # Create a mock failed process
        mock_failed_process = Mock(spec=process.ProductionProcess)
        mock_failed_process.data = Mock()
        mock_failed_process.data.ID = "p1"
        mock_failed_process.get_process_signature.return_value = "ProductionProcess:p1"

        # This should raise a ValueError since no rework processes exist
        with pytest.raises(ValueError, match="No compatible rework processes found"):
            process_matcher_most_trivial.get_rework_compatible(mock_failed_process)

    def test_get_route(self, process_matcher_most_trivial):
        """Test get_route method."""
        # Get locations from the system
        locations = process_matcher_most_trivial._get_all_locations()
        origin = locations[0]
        target = locations[1]

        # Create a mock process
        mock_process = Mock(spec=process.TransportProcess)
        mock_process.get_process_signature.return_value = "TransportProcess:tp"

        # Clear the route cache to ensure no route is found
        process_matcher_most_trivial.route_cache = {}

        # This should raise a ValueError since no route is cached
        with pytest.raises(ValueError, match="No route found"):
            process_matcher_most_trivial.get_route(origin, target, mock_process)

    def test_resource_compatibility_key_creation(self):
        """Test ResourceCompatibilityKey creation and equality."""
        key1 = ResourceCompatibilityKey(process_signature="ProductionProcess:p1")
        key2 = ResourceCompatibilityKey(process_signature="ProductionProcess:p1")
        key3 = ResourceCompatibilityKey(process_signature="ProductionProcess:p2")

        assert key1 == key2
        assert key1 != key3
        assert hash(key1) == hash(key2)
        assert hash(key1) != hash(key3)

    def test_transport_compatibility_key_creation(self):
        """Test TransportCompatibilityKey creation and equality."""
        key1 = TransportCompatibilityKey(
            origin_id="source1",
            target_id="sink1",
            process_signature="TransportProcess:tp",
        )
        key2 = TransportCompatibilityKey(
            origin_id="source1",
            target_id="sink1",
            process_signature="TransportProcess:tp",
        )
        key3 = TransportCompatibilityKey(
            origin_id="source1",
            target_id="sink2",
            process_signature="TransportProcess:tp",
        )

        assert key1 == key2
        assert key1 != key3
        assert hash(key1) == hash(key2)
        assert hash(key1) != hash(key3)

    def test_empty_system_compatibility(self):
        """Test compatibility precomputation with an empty system."""
        # Create an empty production system
        empty_system = psx.ProductionSystem([], [], [])
        adapter = empty_system.to_model()

        from prodsys import runner

        runner_instance = runner.Runner(production_system_data=adapter)
        runner_instance.initialize_simulation()
        process_matcher = (
            runner_instance.product_factory.router.request_handler.process_matcher
        )

        # Clear all compatibility tables
        process_matcher.production_compatibility = {}
        process_matcher.transport_compatibility = {}
        process_matcher.reachability_cache = {}
        process_matcher.rework_compatibility = {}
        process_matcher.route_cache = {}

        # Run precomputation
        process_matcher.precompute_compatibility_tables()

        # Check that tables are empty but initialized
        assert len(process_matcher.production_compatibility) == 0
        assert len(process_matcher.transport_compatibility) == 0
        assert len(process_matcher.reachability_cache) == 0
        assert len(process_matcher.rework_compatibility) == 0
        assert len(process_matcher.route_cache) == 0



from __future__ import annotations

import os
import time

from stable_baselines3 import PPO
from stable_baselines3.common.logger import configure
from stable_baselines3.common.callbacks import BaseCallback

import numpy as np
from gymnasium import spaces

from prodsys.adapters import ProductionSystemData
from prodsys.control import routing_control_env


class RoutingControlEnv(routing_control_env.AbstractRoutingControlEnv):
    def get_observation(self) -> np.ndarray:
        """
        Function that utilizes the ResourceObserver of the environment class to get an array of observations concerning the availability of resources.

        Returns:
            np.ndarray: The observation.
        """
        available_resources = []
        for obs in self.observers:
            resource_available_status = obs.observe_resource_available()
            available_resources.append(resource_available_status.available)

        return np.array(available_resources)

    def get_info(self) -> dict:
        return {"info": 0}

    def get_termination_condition(self) -> bool:
        return self.runner.env.now >= 300000

    def get_reward(self, invalid_action: bool = False) -> float:
        if invalid_action:
            reward = -1
        else:
            reward = 1

        if self.step_count % 10 == 0:
            queue_capacity = sum(
                queue.capacity
                for queue in self.adapter.port_data
                if queue.ID != "SinkQueue"
            )
            resource_capacity = sum(
                resource.capacity for resource in self.adapter.resource_data
            )
            wip = len(self.runner.product_factory.products)
            reward += (queue_capacity + resource_capacity) / wip * 100

        return reward


class TensorboardCallback(BaseCallback):
    """
    Custom callback for plotting additional values in tensorboard.
    """

    def __init__(self, verbose=0):
        super(TensorboardCallback, self).__init__(verbose)

    def _on_step(self) -> bool:
        self.logger.record("reward", self.training_env.get_attr("reward")[0])

        return True


if __name__ == "__main__":
    adapter_object = ProductionSystemData.read(
        "examples/control/control_example_data/control_configuration.json"
    )

    num_of_resources = len(adapter_object.resource_data)

    observation_space = spaces.Box(0, 1, shape=(num_of_resources,), dtype=float)
    action_space = spaces.Box(0, 1, shape=(num_of_resources,), dtype=float)

    env = RoutingControlEnv(
        adapter_object,
        observation_space=observation_space,
        action_space=action_space,
        render_mode="human",
    )

    tmp_path = os.path.join(
        os.getcwd(), "tensorboard_log", "routing", time.strftime("%Y%m%d-%H%M%S")
    )
    new_logger = configure(tmp_path, ["stdout", "csv", "tensorboard"])

    model = PPO(env=env, policy="MlpPolicy", verbose=1)
    model.set_logger(new_logger)
    model.learn(total_timesteps=1000000, callback=TensorboardCallback())

    # Start Tensorboard with: tensorboard --logdir tensorboard_log\routing



import prodsys.express as psx
import prodsys

print("version used:", prodsys.VERSION)
prodsys.set_logging("CRITICAL")

# Create time models
t1 = psx.FunctionTimeModel("normal", 1, 0.1, "t1")
t2 = psx.FunctionTimeModel("normal", 2, 0.2, "t2")

# Create processes
p1 = psx.ProductionProcess(t1, "p1")
p2 = psx.ProductionProcess(t2, "p2")

t3 = psx.DistanceTimeModel(speed=180, reaction_time=0.1, ID="t3")
tp = psx.TransportProcess(t3, "tp")

# Create setup states
s1 = psx.FunctionTimeModel("exponential", 0.5, ID="s1")
setup_state_1 = psx.SetupState(s1, p1, p2, "S1")
setup_state_2 = psx.SetupState(s1, p2, p1, "S2")

# Create resources
machine = psx.Resource(
    [p1, p2],
    [5, 0],
    2,
    states=[setup_state_1, setup_state_2],
    ID="machine",
)
machine2 = psx.Resource(
    [p1, p2],
    [7, 0],
    2,
    states=[setup_state_1, setup_state_2],
    ID="machine2",
)

transport = psx.Resource([tp], [2, 0], 1, ID="transport")

# Create products
product1 = psx.Product(process=[p1, p2], transport_process=tp, ID="product1")
product2 = psx.Product(process=[p2, p1], transport_process=tp, ID="product2")

# Create sinks
sink1 = psx.Sink(product1, [10, 0], "sink1")
sink2 = psx.Sink(product2, [10, 0], "sink2")

# Create orders with different release times
order1 = psx.Order(
    ID="order1",
    ordered_products=[psx.OrderedProduct(product=product1, quantity=2)],
    order_time=0.0,
    release_time=10.0,
    priority=1,
)

order2 = psx.Order(
    ID="order2",
    ordered_products=[psx.OrderedProduct(product=product2, quantity=1)],
    order_time=5.0,
    release_time=15.0,
    priority=1,
)

# Order with multiple product types
order3 = psx.Order(
    ID="order3",
    ordered_products=[
        psx.OrderedProduct(product=product1, quantity=1),
        psx.OrderedProduct(product=product2, quantity=1),
    ],
    order_time=10.0,
    release_time=20.0,
    priority=1,
)

# Create order source
order_source = psx.OrderSource(
    orders=[order1, order2, order3],
    location=[0, 0],
    ID="order_source",
)

# Create production system
system = psx.ProductionSystem(
    [machine, machine2, transport], [order_source], [sink1, sink2]
)

# Set ConWip limit
model = system.to_model()
model.conwip_number = 5

# Run simulation
runner_instance = prodsys.runner.Runner(production_system_data=model)
runner_instance.initialize_simulation()
system.run(100)

runner_instance = system.runner

runner_instance.print_results()
runner_instance.plot_results()
runner_instance.save_results_as_csv()




from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum

from prodsys.simulation import locatable
from prodsys.simulation.dependency import DependedEntity
from typing import Optional, List, Union
from prodsys.simulation.dependency import Dependency
from prodsys.simulation import resources

class Entity(ABC):
    """
    Base class for all entities that flow through the production system.
    
    Entities:
    - Occupy locations (resources, queues, nodes)
    - Can be transported between locations
    - Occupy capacity in resources and queues
    - Can be routed through the system
    """

    @property
    def current_locatable(self) -> locatable.Locatable:
        if self.current_dependant:
            return self.current_dependant.current_locatable
        return self._current_locatable


    @property
    @abstractmethod
    def type(self) -> EntityType:
        """Type of the entity"""
        pass
    
    @property
    @abstractmethod
    def size(self) -> int:
        """Capacity units this entity requires"""
        pass
    
    @abstractmethod
    def update_location(self, locatable: locatable.Locatable) -> None:
        """Update current location"""
        pass

class EntityType(str, Enum):
    PRODUCT = "product"
    PRIMITIVE = "primitive"
    LOT = "lot"


from __future__ import annotations

from typing import Literal, Optional, Union, List
from uuid import uuid1
from pydantic import Field
from pydantic.dataclasses import dataclass

from prodsys.express import core
from prodsys.models.dependency_data import ToolDependencyData, ProcessDependencyData, ResourceDependencyData, LotDependencyData, AssemblyDependencyData, DisassemblyDependencyData
from prodsys.express.node import Node


@dataclass
class Dependency(core.ExpressObject):
    """
    Abstract base class to represent a dependency.
    """

@dataclass
class ResourceDependency(Dependency):
    required_resource: Resource
    interaction_node: Optional[Node]
    position_type: Literal["relative", "absolute"] = "absolute"
    per_lot: bool = True
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))

    def to_model(self):
        # Infer a resource type hint to support downstream dependency reconciliation.
        resource_type_hint = ""
        try:
            processes = getattr(self.required_resource, "processes", []) or []
            if processes:
                # Import locally to avoid circular imports.
                from prodsys.express import process as _process_module

                if any(
                    isinstance(proc, (_process_module.TransportProcess, _process_module.LinkTransportProcess))
                    for proc in processes
                ):
                    resource_type_hint = "transport_resource"
                else:
                    resource_type_hint = "production_resource"
        except Exception:
            # Fallback to empty hint if inference fails.
            resource_type_hint = ""

        return ResourceDependencyData(
            ID=self.ID,
            description=resource_type_hint,
            required_resource=self.required_resource.ID,
            interaction_node=self.interaction_node.ID if self.interaction_node else None,
            position_type=self.position_type,
            per_lot=self.per_lot,
        )


@dataclass
class ProcessDependency(Dependency):
    required_process: PROCESS_UNION
    interaction_node: Optional[Node]
    position_type: Literal["relative", "absolute"] = "absolute"
    per_lot: bool = True
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))

    def to_model(self):
        return ProcessDependencyData(
            ID=self.ID,
            description="",
            required_process=self.required_process.ID,
            interaction_node=self.interaction_node.ID if self.interaction_node else None,
            position_type=self.position_type,
            per_lot=self.per_lot,
        )


@dataclass
class ToolDependency(Dependency):
    required_entity: Union[Product, Primitive]
    per_lot: bool = False
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))

    def to_model(self):
        return ToolDependencyData(
            ID=self.ID,
            description="",
            required_entity=self.required_entity.ID,
            per_lot=self.per_lot,
        )


@dataclass
class AssemblyDependency(Dependency):
    required_entity: Union[Product, Primitive]
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))

    def to_model(self):
        return AssemblyDependencyData(
            ID=self.ID,
            description="",
            required_entity=self.required_entity.ID,
        )


@dataclass
class DisassemblyDependency(Dependency):
    required_entity: Union[Product, Primitive]
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))

    def to_model(self):
        return DisassemblyDependencyData(
            ID=self.ID,
            description="",
            required_entity=self.required_entity.ID,
        )

@dataclass
class LotDependency(Dependency):
    min_lot_size: int
    max_lot_size: int
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))

    def to_model(self):
        return LotDependencyData(
            ID=self.ID,
            description="",
            min_lot_size=self.min_lot_size,
            max_lot_size=self.max_lot_size,
        )
from prodsys.express.primitive import Primitive
from prodsys.express.process import PROCESS_UNION
from prodsys.express.product import Product
from prodsys.express.resources import Resource




from typing import Protocol

from prodsys.models import core_asset

class Locatable(Protocol):
    """
    Protocol that defines the interface for a locatable entity.
    """

    @property
    def data(self) -> core_asset.Locatable: # TODO: maybe make type hints more clear here...
        """
        Returns the data of the locatable.
        """
        pass

    def get_location(self) -> list[float]:
        """
        Returns the location of the locatable.
        """
        pass

    def update_location(self, location: list[float]) -> None:
        """
        Updates the location of the locatable.
        """
        pass


import numpy as np
import skimage
import math
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import matplotlib.patches as mpatches
from scipy.spatial import Voronoi, voronoi_plot_2d
import shapely
from shapely.geometry import Polygon
import pygeoops
from prodsys.util.node_link_generation.configuration import Configuration 


"""
BSD 3-Clause License

Copyright (c) 2023, pygeoops

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


class Station:
    def __init__(self, station_number):
        self.station_number = station_number
        self.station_node = None
        self.station_orientation = None
        self.station_trajectory_nodes = []
        self.station_buffer_nodes = []
        self.station_corner_nodes = []


class StationConfiguration:
    def __init__(self, config, table_config):
        self.table_config = table_config
        self.buffer_node_distance = config.get(Configuration.Buffer_Node_Distance)
        self.stations = []
        self.station_nodes = []
        self.station_orientations = []
        self.station_trajectory_nodes = []
        self.station_buffer_nodes = []

    def add_stations(self, stations: list, visualization) -> None:
        """
        Function adds stations to the station configuration.
        """
        for station_nr, station in enumerate(stations["physical_objects"]):
            # TODO: Implement check, if station is positioned correctly.
            station_object = Station(station_nr)
            self.stations.append(station_object)

            self.station_nodes.append(tuple(station["pose"][:2]))
            self.station_orientations.append(station["pose"][2])
            station_object.station_node = tuple(station["pose"][:2])
            station_object.station_orientation = station["pose"][2]

            trajectory_nodes = self.get_trajectory_node_position(station)
            trajectory_nodes_cfree = self.check_station_nodes_cfree(trajectory_nodes)
            if len(trajectory_nodes_cfree) != len(trajectory_nodes):
                raise ValueError(str(len(trajectory_nodes) - len(trajectory_nodes_cfree)) + " trajectory nodes are not in cfree. \
                      Station number: " + str(station_nr) + " " + str(self.station_nodes[-1]))
            self.station_trajectory_nodes.extend([*trajectory_nodes_cfree])
            station_object.station_trajectory_nodes.extend([*trajectory_nodes_cfree])

            buffer_nodes = self.get_buffer_node_position(station, number_trajectory_nodes=len(trajectory_nodes_cfree),
                                                         buffer_distance=self.buffer_node_distance)
            buffer_nodes = self.check_station_nodes_cfree(buffer_nodes)
            self.station_buffer_nodes.extend([*buffer_nodes])
            station_object.station_buffer_nodes.extend([*buffer_nodes])

            if "stations_layer" not in visualization.occupancy_matrices.keys():
                visualization.add_free_layer("stations_layer")
            station_corners = self.get_station_corner_nodes(station)
            station_object.station_corner_nodes.extend(station_corners)
            visualization.draw_polygon_to_layer("stations_layer", station_corners)

    def remove_all_stations(self, node_edge_generator, visualization):
        """
        Function deletes all stations as well as the corresponding nodes and edges.
        """
        visualization.add_free_layer("stations_layer")
        self.stations = []
        self.station_nodes = []
        self.station_orientations = []
        self.station_trajectory_nodes = []
        self.station_buffer_nodes = []
        node_edge_generator.graph.remove_station_nodes_and_edges()

    # Check nodes and edges.
    def check_station_nodes_cfree(self, nodes: list) -> list:
        """
        Check if nodes are in cfree.
        """
        nodes_in_cfree = []
        for node in nodes:
            point = shapely.Point(node)
            if point.within(self.table_config.table_configuration_polygon_cfree):
                nodes_in_cfree.append(node)
        return nodes_in_cfree

    def get_trajectory_node_position(self, station: tuple) -> list:
        """
        Function returns the trajectory nodes of a station. Stations are quadratic.
        """
        # Define local trajectory nodes relative to the station center node.
        trajectory_in_node = (station["poi"]["in_port.trajectory_node"][0], station["poi"]["in_port.trajectory_node"][1])
        trajectory_out_node = (station["poi"]["out_port.trajectory_node"][0], station["poi"]["out_port.trajectory_node"][1])
        if trajectory_in_node == trajectory_out_node:
            local_trajectory_nodes = [trajectory_in_node]
        elif trajectory_in_node != trajectory_out_node:
            local_trajectory_nodes = [trajectory_in_node, trajectory_out_node]

        trajectory_nodes = self.rotate_stations(station, local_trajectory_nodes)

        return trajectory_nodes

    def get_buffer_node_position(self, station: tuple, number_trajectory_nodes: int, buffer_distance: int) -> list:
        """
        Function returns the buffer nodes of a station. Stations are quadratic.
        """
        # Define local buffer nodes relative to the station center node.
        if number_trajectory_nodes == 1:
            local_buffer_nodes = [(station["poi"]["in_port.trajectory_node"][0], - buffer_distance),
                                  (station["poi"]["in_port.trajectory_node"][0], buffer_distance)]
            #                      (station["poi"]["in_port.trajectory_node"][0] + buffer_distance, 0),
            #                      (station["poi"]["in_port.trajectory_node"][0] + buffer_distance, - buffer_distance),
            #                      (station["poi"]["in_port.trajectory_node"][0] + buffer_distance, buffer_distance)]

        elif number_trajectory_nodes == 2:
            local_buffer_nodes = [(station["poi"]["in_port.trajectory_node"][0], - buffer_distance),
                                  (station["poi"]["in_port.trajectory_node"][0], buffer_distance),
                                  (station["poi"]["out_port.trajectory_node"][0], - buffer_distance),
                                  (station["poi"]["out_port.trajectory_node"][0], buffer_distance)]

        buffer_nodes = self.rotate_stations(station, local_buffer_nodes)

        return buffer_nodes

    def get_station_corner_nodes(self, station: tuple) -> list:
        """
        Function returns the corner nodes of a station. Stations are quadratic.
        """
        # Define the corners of the station in its local coordinate system.
        local_corners = [(station["bounding_box"][0][0], station["bounding_box"][0][1]),
                         (station["bounding_box"][1][0], station["bounding_box"][0][1]),
                         (station["bounding_box"][1][0], station["bounding_box"][1][1]),
                         (station["bounding_box"][0][0], station["bounding_box"][1][1])]

        station_corners = self.rotate_stations(station, local_corners)

        return station_corners

    def rotate_stations(self, station, local_nodes) -> list:
        """
        Function calculates the global coordinates of local station nodes of rotated stations.
        """
        # Convert the orientation to radians.
        theta = math.radians(station["pose"][2])

        # Define the rotation matrix.
        rotation_matrix = [[math.cos(theta), -math.sin(theta)], [math.sin(theta), math.cos(theta)]]

        # Rotate the corners and translate them to the global coordinate system.
        station_corners = [(round(station["pose"][0] + rotation_matrix[0][0] * node[0] + rotation_matrix[0][1] * node[1]),
                            round(station["pose"][1] + rotation_matrix[1][0] * node[0] + rotation_matrix[1][1] * node[1]))
                           for node in local_nodes]

        return station_corners


class Table:
    def __init__(self, table_number=None):
        self.table_number = table_number
        self.table_orientation = None
        self.table_corner_nodes = []
        self.table_configuration_intersection_middle_offset_nodes = []
        self.table_polygon = None


class TableConfiguration:
    def __init__(self, config):
        # Boundary conditions.
        self.boundary_distance = config.get(Configuration.Boundary_Distance)
        self.required_node_clearance = config.get(Configuration.Min_Node_Distance)
        self.required_node_edge_clearance = config.get(Configuration.Min_Node_Edge_Distance)

        # Table configuration.
        self.tables = []
        self.table_configuration_intersection_middle_nodes = []
        self.table_configuration_corner_nodes = []
        self.table_configuration_corner_nodes_cfree = []
        self.table_configuration_polygon = []
        self.table_configuration_polygon_cfree = []
        self.table_configuration_polygon_cfree_round = []
        self.table_configuration_polygon_without_stations = []

        # Medial axis.
        self.medial_axis = None
        self.medial_axis_nodes = []
        self.medial_axis_edges = []

        # Zones.
        self.zones_polygons = []
        self.zones_directions = []
        self.zones_edges = []

    def reset_table_configuration(self, visualization, update=False):
        """
        Function is used to reset the table configuration.
        """
        # Table configuration.
        if update:
            # Reset table configuration polygon. Table configuration polygon is the union of all table modules without the stations.
            self.table_configuration_polygon = self.table_configuration_polygon_without_stations
        else:
            # Reset table configuration completely.
            visualization.add_table_layer()
            self.tables = []
            self.table_configuration_intersection_middle_nodes = []
            self.table_configuration_corner_nodes = []
            self.table_configuration_corner_nodes_cfree = []
            self.table_configuration_polygon = []
            self.table_configuration_polygon_cfree = []
            self.table_configuration_polygon_cfree_round = []
            self.table_configuration_polygon_without_stations = []

        # Medial axis.
        self.medial_axis = None
        self.medial_axis_nodes = []
        self.medial_axis_edges = []

        # Zones.
        self.zones_polygons = []
        self.zones_directions = []
        self.zones_edges = []

    def add_table(self, table_corners_poses, visualization, layer_id="table_layer") -> None: #MARKER
        """
        Function adds table modules. Table layer is adapted for visualization.
        table_corners_poses: List of poses (x, y, theta) of the corners of the table module.
                            x, y are the coordinates of the corner. theta is the orientation of the table module.
        visualization: Visualization object.
        layer_id: Layer id for visualization.
        """
        if "table_layer" not in visualization.occupancy_matrices.keys():
            # Add a table layer if no table layer exists yet.
            visualization.add_table_layer()

        # Initialize table object.
        table_object = Table()
        self.tables.append(table_object)

        # Extract the corner nodes of the table module.
        table_corner_nodes = []
        for corner in table_corners_poses:
            table_corner_nodes.append((corner['pose'][0], corner['pose'][1]))

        # Define table module as polygon.
        table_polygon = shapely.Polygon(table_corner_nodes)

        if self.table_configuration_polygon == []:
            # No table module added so far.
            self.table_configuration_polygon = table_polygon

        # Check if table module is overlapping with the existing table modules. #MARKER kann man das entfernen oder wird es sonst buggy?
        elif not table_polygon.contains(self.table_configuration_polygon):# and not table_polygon.within(self.table_configuration_polygon):
            # Define table polygon of table object for calculate_intersection_nodes.
            table_object.table_polygon = table_polygon
            # Check if the polygons intersect.
            if table_polygon.intersects(self.table_configuration_polygon):
                # Extract the intersection line(s) of the two polygons.
                intersection_line = table_polygon.intersection(self.table_configuration_polygon)
                self.calculate_intersection_nodes(intersection_line)

            # Define table configuration polygon.
            self.table_configuration_polygon = shapely.unary_union([table_polygon, self.table_configuration_polygon])

        else:
            raise ValueError("Table module is overlapping with the existing table modules or is within the existing table configuration: " + str(table_corner_nodes))
        
        # Table module is not overlapping with the existing table modules and can be added.
        table_object.table_orientation = table_corners_poses[0]['pose'][2]
        table_object.table_polygon = table_polygon
        table_object.table_corner_nodes = table_corner_nodes

        # Add table module for visualization.
        # TODO: How to handle case when table is not parallel to the axes or table has more than 4 corners?
        visualization.occupancy_matrices[layer_id][int(table_corner_nodes[0][1]):int(table_corner_nodes[2][1]) + 1,
                                                    int(table_corner_nodes[0][0]):int(table_corner_nodes[2][0]) + 1] = 0

    def remove_table(self, table_corner_nodes, visualization, layer_id="table_layer") -> None:
        """
        Function removes table modules. Table layer is adapted for visualization.
        table_corner_nodes: Corner nodes of the table modules. List of (x, y) coordinates.
        visualization: Visualization object.
        layer_id: Layer id for visualization.
        """
        # TODO: Visualization is not correct after table modules are removed. Reset table configuration and add all tables again is easier.
        # Define table module as polygon.
        table_polygon = shapely.Polygon(table_corner_nodes)

        # Check if table module exists.
        remove_table = False
        for table in self.tables:
            if table.table_polygon == table_polygon:
                remove_table = True
                table_object = table

        if not remove_table:
            raise ValueError("Table module does not exist.")

        # Table module exists and can be removed.
        self.table_configuration_polygon = self.table_configuration_polygon.difference(table_polygon)
        self.tables.remove(table_object)

        # Remove table module for visualization.
        # TODO: How to handle case when table is not parallel to the axes or table has more than 4 corners?
        visualization.occupancy_matrices[layer_id][table_corner_nodes[0][1]:table_corner_nodes[2][1] + 1,
                                                   table_corner_nodes[0][0]:table_corner_nodes[2][0] + 1] = 1

    def generate_table_configuration(self) -> None:
        """
        Function extracts the corner nodes and the usable corner nodes for the vehicles of the table configuration.
        """
        if self.table_configuration_corner_nodes != [] and self.table_configuration_corner_nodes_cfree != []:
            # Reset table configuration corner nodes.
            self.table_configuration_corner_nodes = []
            self.table_configuration_corner_nodes_cfree = []

        # Remove intermediate nodes of the polygon along an edge.
        # Now interior and exterior nodes of the polygon are only corner nodes of the polygon.
        self.table_configuration_polygon = self.table_configuration_polygon.buffer(-1, join_style=2)
        self.table_configuration_polygon = self.table_configuration_polygon.buffer(+1, join_style=2)

        # Store a version of the table configuration polygon, which remains unchanged when stations are added.
        self.table_configuration_polygon_without_stations = self.table_configuration_polygon.buffer(-1, join_style=2)
        self.table_configuration_polygon_without_stations = self.table_configuration_polygon_without_stations.buffer(+1, join_style=2)

        # Define table configuration corner nodes. Nodes should lie within the table configuration polygon, not at the outer edge.
        self.table_configuration_corner_nodes = self.update_table_configuration_corner_nodes(
            self.table_configuration_polygon.buffer(- 0.01, join_style=2))

        # Define usable space for vehicles.
        self.table_configuration_polygon_cfree = self.table_configuration_polygon.buffer(0.01 - self.boundary_distance, join_style=2)
        self.table_configuration_polygon_cfree_round = self.table_configuration_polygon.buffer(0.01 - self.boundary_distance)

        # Define usable corner nodes of the table configuration for the vehicles.
        # Nodes should lie within the table configuration cfree polygon, not at the outer edge.
        self.table_configuration_corner_nodes_cfree = self.update_table_configuration_corner_nodes(
            self.table_configuration_polygon.buffer(- self.boundary_distance, join_style=2))

    def table_configuration_with_stations(self, station_config):
        """
        Function extracts the corner nodes and the usable corner nodes for the vehicles
        of the table configuration considering the position of the stations.
        """
        # For all stations, check if they are positioned (partly) on the table configuration.
        for station in station_config.stations:
            station_area = shapely.Polygon(station.station_corner_nodes)
            if station_area.intersects(self.table_configuration_polygon):
                # Station is (partly) on the table configuration. Adjust the polygon of the table configuration.
                self.table_configuration_polygon = self.table_configuration_polygon.difference(station_area)

        # Define the usable space as well as the table configuration corner nodes and usable corner nodes for the vehicles.
        # Nodes should lie within the table configuration polygons, not at the outer edge.
        self.table_configuration_polygon_cfree = self.table_configuration_polygon.buffer(0.01 - self.boundary_distance, join_style=2)
        self.table_configuration_polygon_cfree_round = self.table_configuration_polygon.buffer(0.01 - self.boundary_distance)
        self.table_configuration_corner_nodes = self.update_table_configuration_corner_nodes(
            self.table_configuration_polygon.buffer(- 0.01, join_style=2))
        self.table_configuration_corner_nodes_cfree = self.update_table_configuration_corner_nodes(
            self.table_configuration_polygon.buffer(- self.boundary_distance, join_style=2))

    def update_table_configuration_corner_nodes(self, polygon) -> list:
        """
        Function orders the table configuration corner nodes of a given polygon.
        Lists of ordered corner nodes starts with the lower left corner node and continues counter-clockwise.
        - [0]: Ordered list of corner nodes of exterior.
        - [1]: Ordered list of corner nodes of first interior (if existing).
        - [2]: Ordered list of corner nodes of second interior (if existing).
        - [3]: ...
        return: List of ordered lists of table configuration corner nodes.
        """
        corner_nodes = []
        ordered_corner_nodes = []

        # Extract exterior corner nodes.
        x, y = polygon.exterior.xy
        for i in range(len(x)):
            if len(corner_nodes) == 0:
                corner_nodes.append([])
                corner_nodes[0].append((x[i], y[i]))
            else:
                corner_nodes[0].append((x[i], y[i]))

        # Extract interior corner nodes. Interior corner nodes only exist, if polygon has a hole/holes in the middle.
        for index in range(len(polygon.interiors)):
            x, y = polygon.interiors[index].xy
            for i in range(len(x)):
                # TODO: Remove intermediate nodes along an edge.
                if len(corner_nodes) == index + 1:
                    corner_nodes.append([])
                    corner_nodes[index + 1].append((x[i], y[i]))
                else:
                    corner_nodes[index + 1].append((x[i], y[i]))

        # Reorder corner nodes. Start at left lower corner and ordered counter-clockwise.
        ordered_corner_nodes = corner_nodes
        for loop_index, corner_nodes_loop in enumerate(corner_nodes):
            corner_nodes_loop.pop(-1)
            start_index = corner_nodes_loop.index(min(corner_nodes_loop))
            part1 = corner_nodes_loop[start_index:]
            part2 = corner_nodes_loop[:start_index]
            ordered_corner_nodes[loop_index] = part1 + part2
            if ordered_corner_nodes[loop_index][0][0] == ordered_corner_nodes[loop_index][1][0]:
                # Corner nodes 1 and 2 have the same x value. Corner nodes are ordered clockwise.
                # Direction must be changed to counter-clockwise.
                ordered_corner_nodes[loop_index].append(ordered_corner_nodes[loop_index][0])
                ordered_corner_nodes[loop_index].reverse()
                ordered_corner_nodes[loop_index].pop(-1)
        return ordered_corner_nodes

    def check_corner_type(self, corner_node) -> str:
        """
        Function checks the type of a corner node of the table configuration.
        return: Type of the corner node. 'convex' or 'concave'.
        """
        for i in [0, 1]:
            if i == 0:
                # Check diagonal neighbor nodes of the corner node.
                neighbor_nodes = [(corner_node[0] - 1, corner_node[1] - 1), (corner_node[0] + 1, corner_node[1] - 1),
                                  (corner_node[0] + 1, corner_node[1] + 1), (corner_node[0] - 1, corner_node[1] + 1)]
            elif i == 1:
                # Check vertical and horizontal neighbor nodes of the corner node.
                neighbor_nodes = [(corner_node[0] - 1, corner_node[1]), (corner_node[0] + 1, corner_node[1]),
                                  (corner_node[0], corner_node[1] - 1), (corner_node[0], corner_node[1] + 1)]
            neighbor_points = [shapely.Point(neighbor_node) for neighbor_node in neighbor_nodes]

            neighbor_points_in_cfree = 0
            for neighbor_point in neighbor_points:
                if neighbor_point.within(self.table_configuration_polygon_cfree):
                    neighbor_points_in_cfree += 1

            if neighbor_points_in_cfree == 3:
                corner_type = 'convex'
                return corner_type
            elif neighbor_points_in_cfree == 1:
                corner_type = 'concave'
                return corner_type
            else:
                corner_type = 'undefined'

        return corner_type

    def define_table_nodes(self, table) -> list:
        """
        Function defines nodes on predefined positions on the table module.
        On each table module, 3 nodes are defined. One node is positioned in the middle of the table module.
        The other two nodes are positioned at the left and right side of the table module.
        The distance between the middle node and the left and right node is 33 cm.
        Distance is defined for vehicles that have a rotation diameter of at most 33 cm.
        """
        table_nodes = []

        # Define table nodes based on the given table module.
        table_nodes.append((table.table_polygon.centroid.x, table.table_polygon.centroid.y))
        if table.table_orientation == 0:
            table_nodes.append((table.table_polygon.centroid.x - 33, table.table_polygon.centroid.y))
            table_nodes.append((table.table_polygon.centroid.x + 33, table.table_polygon.centroid.y))
        elif table.table_orientation == 90:
            table_nodes.append((table.table_polygon.centroid.x, table.table_polygon.centroid.y - 33))
            table_nodes.append((table.table_polygon.centroid.x, table.table_polygon.centroid.y + 33))

        return table_nodes

    def calculate_intersection_nodes(self, intersection_line, parallel_offset=10) -> None:
        """
        Function calculates the intersection nodes of table modules based on the intersection line.
        """
        # Check if the intersection line is a line or a multi line.
        if isinstance(intersection_line, shapely.LineString):
            # Calculate the middle point of the line.
            mid_point = intersection_line.interpolate(0.5, normalized=True)
            self.table_configuration_intersection_middle_nodes.append((mid_point.x, mid_point.y))

            # Define nodes at +parallel_offset and -parallel_offset distance from the middle point vertically to the intersection line.
            offset_line_positive = intersection_line.parallel_offset(parallel_offset, 'left')
            offset_line_negative = intersection_line.parallel_offset(parallel_offset, 'right')

            mid_point_positive = offset_line_positive.interpolate(0.5, normalized=True)
            mid_point_negative = offset_line_negative.interpolate(0.5, normalized=True)
            mid_point_positive_shapely = shapely.Point(mid_point_positive.x, mid_point_positive.y)
            mid_point_negative_shapely = shapely.Point(mid_point_negative.x, mid_point_negative.y)

            # Check to which table the nodes belong.
            for table in self.tables:
                if mid_point_positive_shapely.within(table.table_polygon):
                    table.table_configuration_intersection_middle_offset_nodes.append((mid_point_positive.x, mid_point_positive.y))
                if mid_point_negative_shapely.within(table.table_polygon):
                    table.table_configuration_intersection_middle_offset_nodes.append((mid_point_negative.x, mid_point_negative.y))

        elif isinstance(intersection_line, shapely.MultiLineString):
            # Calculate the middle point of the lines.
            for line in intersection_line.geoms:
                mid_point = line.interpolate(0.5, normalized=True)
                self.table_configuration_intersection_middle_nodes.append((mid_point.x, mid_point.y))

                # Define nodes at +parallel_offset and -parallel_offset distance from the middle point vertically to the intersection line.
                offset_line_positive = line.parallel_offset(parallel_offset, 'left')
                offset_line_negative = line.parallel_offset(parallel_offset, 'right')

                mid_point_positive = offset_line_positive.interpolate(0.5, normalized=True)
                mid_point_negative = offset_line_negative.interpolate(0.5, normalized=True)
                mid_point_positive_shapely = shapely.Point(mid_point_positive.x, mid_point_positive.y)
                mid_point_negative_shapely = shapely.Point(mid_point_negative.x, mid_point_negative.y)

                # Check to which table the nodes belong.
                for table in self.tables:
                    if mid_point_positive_shapely.within(table.table_polygon):
                        table.table_configuration_intersection_middle_offset_nodes.append((mid_point_positive.x, mid_point_positive.y))
                    if mid_point_negative_shapely.within(table.table_polygon):
                        table.table_configuration_intersection_middle_offset_nodes.append((mid_point_negative.x, mid_point_negative.y))

    def define_medial_axis(self) -> None:
        """
        Function defines the medial axis of the table configuration considering the stations.
        """
        # Define medial axis of the table configuration.
        self.medial_axis = pygeoops.centerline(self.table_configuration_polygon_cfree,
                                               densify_distance=-0.01, min_branch_length=0, simplifytolerance=-0.05)

        # Extract medial axis nodes and edges. Iterate over each LineString in the MultiLineString.
        for line in self.medial_axis.geoms:
            # Extract nodes from the LineString.
            line_nodes = list(line.coords)

            # Check distance to the table configuration corner nodes.
            corner_branch = False
            for line_node in line_nodes:
                # Calculate the distance to each corner node and find the minimum distance.
                for corner_nodes in self.table_configuration_corner_nodes_cfree:
                    distances = [math.dist(line_node, corner_node) for corner_node in corner_nodes]
                    min_distance = min(distances)

                    if min_distance < 4:
                        # LineString is too close to a corner node.
                        corner_branch = True
                        break

            if corner_branch:
                # Nodes and edges are not added, if it is a corner branch.
                continue

            # Add nodes to the medial axis nodes.
            self.medial_axis_nodes.extend(line_nodes)

            # Extract edges from the LineString and add them to the medial axis edges.
            line_edges = [(line_nodes[i], line_nodes[i+1]) for i in range(len(line_nodes) - 1)]
            self.medial_axis_edges.extend(line_edges)

        # Remove duplicate nodes.
        self.medial_axis_nodes = list(dict.fromkeys(self.medial_axis_nodes))

        # Define simplified medial axis without the corner branches.
        self.medial_axis = shapely.geometry.MultiLineString(self.medial_axis_edges)

    def define_zones(self, layout_nr=None) -> None:
        """
        Function defines zones of the table configuration.
        """
        corner_nodes_points = []
        corner_nodes_cfree = []
        corner_nodes_cfree_points = []

        # Add edges between the corner nodes of the table configuration (boundary of the table configuration).
        for nodes in self.table_configuration_corner_nodes:
            corner_nodes_points.extend([shapely.Point(node) for node in nodes])
            for node_index, corner_node in enumerate(nodes):
                if node_index == len(nodes) - 1:
                    self.zones_edges.append(shapely.LineString([corner_node, nodes[0]]))
                else:
                    self.zones_edges.append(shapely.LineString([corner_node, nodes[node_index + 1]]))

        # Define corner nodes in cfree.
        for nodes in self.table_configuration_corner_nodes_cfree:
            corner_nodes_cfree.extend(nodes)
            corner_nodes_cfree_points.extend([shapely.Point(node) for node in nodes])

        # Add edges between the corner nodes in cfree of the table configuration and the closest medial axis nodes.
        # In addition add edges between the corner nodes in cfree and the closest corner nodes of the table configuration.
        for corner_node_cfree in corner_nodes_cfree:
            # Calculate the distance to each medial axis node and find the minimum distance.
            distances = [math.dist(corner_node_cfree, medial_axis_node) for medial_axis_node in self.medial_axis_nodes]
            min_distance = min(distances)

            # Find the medial axis node with the minimum distance.
            min_distance_index = distances.index(min_distance)
            min_distance_node = self.medial_axis_nodes[min_distance_index]

            # Calculate the distance to each corner node and find the minimum distance. Exclude distance == 0.
            line = shapely.LineString([corner_node_cfree, min_distance_node])
            node_distances = line.distance(corner_nodes_cfree_points)
            min_node_distance = min([distance for distance in node_distances if distance != 0])

            # Add edge, if distance to corner nodes is large enough.
            if min_node_distance > 10:
                self.zones_edges.append(shapely.LineString([corner_node_cfree, min_distance_node]))

                # Find closest corner node and add edge.
                closest_corner_node = min(corner_nodes_points, key=lambda node: node.distance(shapely.Point(corner_node_cfree)))
                self.zones_edges.append(shapely.LineString([closest_corner_node, corner_node_cfree]))

        # Add medial axis edges.
        for line in self.medial_axis.geoms:
            self.zones_edges.append(line)

        # Create a MultiLineString object.
        self.zones_edges = shapely.geometry.MultiLineString(self.zones_edges)

        # Define polygons based on the zones edges.
        polygons = shapely.ops.polygonize(self.zones_edges)
        self.zones_polygons = list(polygons)

        # Remove polygons that are not within the table configuration.
        for polygon in self.zones_polygons:
            if not polygon.within(self.table_configuration_polygon):
                self.zones_polygons.remove(polygon)

        self.define_zones_directions(layout_nr)

    def define_zones_directions(self, layout_nr) -> None:
        """
        Function is used to define prioritized movement directions for the zones.
        Directions are right, left, up and down.
        """
        # TODO: Implement algorithm. Zones are currently hard coded for few layouts.
        for zone in self.zones_polygons:
            pass

        if layout_nr == 12:
            # Zones are defined counter-clockwise following the outer boundary of the table configuration. No inner boundary.
            self.zones_directions = ['right', 'up', 'left', 'up', 'left', 'down', 'left', 'down']
        elif layout_nr == 15:
            # Zones are defined clockwise following the outer boundary of the table configuration.
            # At the inner boundary, zones are defined counter-clockwise.
            self.zones_directions = ['left', 'down', 'right', 'up', 'right', 'up', 'left', 'down']
        else:
            self.zones_directions = []


class Visualization:
    def __init__(self, config, table_config, station_config, node_edge_generator):
        # Objects
        self.config = config
        self.table_config = table_config
        self.station_config = station_config
        self.node_edge_generator = node_edge_generator

        # Occupancy matrices.
        self.occupancy_matrices = {}

    def add_free_layer(self, layer_id: str) -> None:
        """
        Function adds a free layer (value 0 -> usable space) of the given layer_id. Layer is only used for visualization.
        """
        self.occupancy_matrices[layer_id] = np.zeros((self.config.get(Configuration.Dim_Y), self.config.get(Configuration.Dim_X)))

    def add_table_layer(self, layer_id="table_layer") -> None:
        """
        Function adds a table layer (value 1 -> non-usable space). Table modules are added separately (value 0 -> usable space).
        Layer is only used for visualization.
        """
        self.occupancy_matrices[layer_id] = np.ones((self.config.get(Configuration.Dim_Y), self.config.get(Configuration.Dim_X))) \
            * self.config.get(Configuration.Blocked_Space_Value)
        self.occupancy_matrices[layer_id][0, 0] = 1

    def add_dilation_difference_layer(self, value: float = 0.6, kernel_radius: int = 0, layer_id: str = 'dilation_difference_layer') -> None:
        """
        Function adds a dilation difference layer (expanded non-usable space). Layer is only used for visualization.
        """
        self.occupancy_matrices[layer_id] = np.zeros((self.config.get(Configuration.Dim_Y), self.config.get(Configuration.Dim_X)))

        dilated_layers = self.get_dialated_occupancy_layers(None, self.config.get(Configuration.Boundary_Distance))
        occupancy_layers = self.get_merged_occupancy_layers(None)

        self.occupancy_matrices[layer_id] = (dilated_layers - occupancy_layers) * self.config.get(Configuration.Blocked_Space_Value) * value

    def calculate_padding(self, current_shape, new_shape):
        """
        Function calculates the padding needed for the current layer to match the size of the new layer.
        """
        padding_y = (0, new_shape[0] - current_shape[0]) if new_shape[0] > current_shape[0] else (0, 0)
        padding_x = (0, new_shape[1] - current_shape[1]) if new_shape[1] > current_shape[1] else (0, 0)

        return padding_y, padding_x

    def adjust_layer_dimensions(self) -> None:
        """
        Function adjusts the dimensions of the layers to the dimensions of the production layout.
        """
        # Remove dilation difference layer.
        if 'dilation_difference_layer' in self.occupancy_matrices.keys():
            self.occupancy_matrices.pop('dilation_difference_layer')

        # Adjust the dimensions of the layers to the dimensions of the production layout.
        for layer_id, current_layer in self.occupancy_matrices.items():
            if layer_id == "table_layer":
                new_layer = np.ones((self.config.get(Configuration.Dim_Y), self.config.get(Configuration.Dim_X)))
                fill_value = self.config.get(Configuration.Blocked_Space_Value)
            elif layer_id == "stations_layer":
                new_layer = np.zeros((self.config.get(Configuration.Dim_Y), self.config.get(Configuration.Dim_X)))
                fill_value = 0
            else:
                continue

            # Calculate the padding needed for the current layer
            padding_y, padding_x = self.calculate_padding(current_layer.shape, new_layer.shape)

            # If current_layer is smaller in y direction, pad it to match the size of new_layer
            if current_layer.shape[0] < new_layer.shape[0]:
                current_layer = np.pad(current_layer, (padding_y, (0, 0)), mode='constant', constant_values=fill_value)
            # If current_layer is larger in y direction, slice it to match the size of new_layer
            elif current_layer.shape[0] > new_layer.shape[0]:
                current_layer = current_layer[:new_layer.shape[0], :]

            # If current_layer is smaller in x direction, pad it to match the size of new_layer
            if current_layer.shape[1] < new_layer.shape[1]:
                current_layer = np.pad(current_layer, ((0, 0), padding_x), mode='constant', constant_values=fill_value)
            # If current_layer is larger in x direction, slice it to match the size of new_layer
            elif current_layer.shape[1] > new_layer.shape[1]:
                current_layer = current_layer[:, :new_layer.shape[1]]

            # Store the adjusted layer
            self.occupancy_matrices[layer_id] = current_layer

    def get_dialated_occupancy_layers(self, layer_ids: str = None, kernel_radius: int = 0) -> np.array:

        dialated_occupancy_layer = self.get_merged_occupancy_layers(layer_ids)

        if kernel_radius != 0:
            dialated_occupancy_layer = skimage.morphology.isotropic_dilation(dialated_occupancy_layer, kernel_radius, out=None, spacing=None)

        return dialated_occupancy_layer

    def get_merged_occupancy_layers(self, ids: list = None) -> np.array:

        merged_occupancy_layers = self.get_occupancy_layers(ids)
        stacked_occupancy_matrices = np.maximum.reduce(merged_occupancy_layers)

        return stacked_occupancy_matrices

    def get_occupancy_layer(self, layer_id: str) -> np.array:

        return self.occupancy_matrices[layer_id]

    def get_occupancy_layers(self, layer_ids: str = None) -> list:

        occupancy_matrices = []

        if layer_ids is None:
            layer_ids = list(self.occupancy_matrices.keys())

        for id in layer_ids:
            occupancy_matrices.append(self.get_occupancy_layer(id))

        return occupancy_matrices

    def get_occupied_coordinates(self, ids: list = None) -> np.array:
        # TODO: Function is not used.

        coordinates = np.nonzero(self.get_merged_occupancy_layers(ids))

        return coordinates

    def get_unoccupied_cooridnates_sampled(self, sampling_distance: int, dialation_radius: int = 0, ids: list = None) -> list:
        # TODO: Function is not used.

        sample_grid = self.get_dialated_occupancy_layers(ids, dialation_radius)
        samples = []

        start_x = int(np.min(self.table_configuration_corner_nodes[0][:], 0)[0])
        start_y = int(np.min(self.table_configuration_corner_nodes[0][:], 0)[1])

        for x in range(start_x + dialation_radius + 1, self.config.get(Configuration.Dim_X), sampling_distance):
            for y in range(start_y + dialation_radius + 1, self.config.get(Configuration.Dim_Y), sampling_distance):
                if not sample_grid[y, x]:
                    samples.append((x, y))

        return samples

    def get_unoccupied_coordinates(self, ids: list = None) -> np.array:
        # TODO: Function is not used.

        coordinates = np.nonzero(~self.get_merged_occupancy_layers(ids))

        return coordinates

    def remove_occupancy_layer(self, id) -> None:

        self.occupancy_matrices.pop(id)

    def reset_stack(self) -> None:
        self.occupancy_matrices = {}

    def set_occupancy_layer(self, layer_id: str, layer_matrix: np.array) -> None:

        self.occupancy_matrices[layer_id] = layer_matrix

    def draw_ellipse_to_layer(self, layer_id: str, center: tuple, rad_x: int, rad_y: int, rot: float) -> None:

        rr, cc = skimage.draw.ellipse(center[1], center[0], rad_y, rad_x,
                                        shape=(self.config.get(Configuration.Dim_Y), self.config.get(Configuration.Dim_X)), rotation=-rot)
        self.occupancy_matrices[layer_id][rr, cc] = self.config.get(Configuration.Blocked_Space_Value)

    def draw_polygon_to_layer(self, layer_id: str, vertices: list, value=1) -> None:
        """
        The function draws a polygon based on the given vertices. Polygons are only used for visualization.
        """

        lst = list(zip(*vertices))

        vertices_x = lst[0]
        vertices_y = lst[1]

        rr, cc = skimage.draw.polygon(vertices_y, vertices_x,
                                        shape=(self.config.get(Configuration.Dim_Y), self.config.get(Configuration.Dim_X)))

        if value == 0:
            factor = 0
        else:
            factor = self.config.get(Configuration.Blocked_Space_Value) / value

        self.occupancy_matrices[layer_id][rr, cc] = factor
        pass

    def show_table_configuration(self, nodes: bool = False, edges: bool = False, tables: bool = False,
                                stations: bool = False, station_nodes: bool = False, table_configuration: bool = False,
                                boundary: bool = False, medial_axis: bool = False, zones: bool = False,
                                layer_ids: list = None, block: bool = True) -> None:
        """
        Function shows the table configuration.
        """
        handles = []
        plt.figure(figsize=(12, 8))
        if tables:
            # Table corner nodes.
            for index, table in enumerate(self.table_config.tables):
                if index == 0:
                    x, y = zip(*table.table_corner_nodes)
                    # plt.scatter(x, y, c="green", marker=".", label="Module")
                    modules,  = plt.plot([x[0], x[1], x[2], x[3], x[0]], [y[0], y[1], y[2], y[3], y[0]],
                                                        linewidth=1.5, c="green", linestyle="-", label="Module")
                else:
                    x, y = zip(*table.table_corner_nodes)
                    # plt.scatter(x, y, c="green", marker=".")
                    plt.plot([x[0], x[1], x[2], x[3], x[0]], [y[0], y[1], y[2], y[3], y[0]], linewidth=1.5, c="green", linestyle="-")

        if table_configuration:
            # Table configuration corner nodes.
            all_corner_nodes = []
            for corner_nodes in self.table_config.table_configuration_corner_nodes:
                all_corner_nodes.extend(corner_nodes)
            x, y = zip(*all_corner_nodes)
            plt.scatter(x, y, c="green", marker="h", label="table configuration corner nodes")

            # Table configuration corner nodes in cfree.
            all_corner_nodes = []
            for corner_nodes in self.table_config.table_configuration_corner_nodes_cfree:
                all_corner_nodes.extend(corner_nodes)
            x, y = zip(*all_corner_nodes)
            plt.scatter(x, y, c="green", marker="x", label="table configuration corner nodes cfree")

        if boundary:
            # x, y = self.table_config.table_configuration_polygon.exterior.xy
            # table_boundary, = plt.plot(x, y, linewidth=3, marker='', color='green', linestyle='-', label='Grenzelinie Hindernisse')
            # for index in range(len(self.table_config.table_configuration_polygon.interiors)):
            #     x, y = self.table_config.table_configuration_polygon.interiors[index].xy
            #     plt.plot(x, y, linewidth=3, marker='', color='green', linestyle='-')

            x, y = self.table_config.table_configuration_polygon_cfree_round.exterior.xy
            cfree, = plt.plot(x, y, linewidth=3, marker='', color='green', linestyle='--', label='Grenzelinie Freifläche')
            for index in range(len(self.table_config.table_configuration_polygon_cfree_round.interiors)):
                x, y = self.table_config.table_configuration_polygon_cfree_round.interiors[index].xy
                plt.plot(x, y, linewidth=3, marker='', color='green', linestyle='--')

        if edges:
            for index, edge in enumerate(self.node_edge_generator.graph.edges):
                x_coords = (edge.node1.position[0], edge.node2.position[0])
                y_coords = (edge.node1.position[1], edge.node2.position[1])
                # x_coords, y_coords = zip(*edge)
                if index == 0:
                    edges_plot, = plt.plot(x_coords, y_coords, color='blue', linestyle='--', linewidth=1.5, label='Kanten')
                    # plt.arrow(x_coords[0], y_coords[0], (x_coords[1] - x_coords[0])/10, (y_coords[1] - y_coords[0])/10, head_width=5, color='blue', linestyle='--')
                    # plt.arrow(x_coords[1], y_coords[1], (x_coords[0] - x_coords[1])/10, (y_coords[0] - y_coords[1])/10, head_width=5, color='blue', linestyle='--')
                else:
                    plt.plot(x_coords, y_coords, color='blue', linestyle='--', linewidth=1.5)
                    # plt.arrow(x_coords[0], y_coords[0], (x_coords[1] - x_coords[0])/10, (y_coords[1] - y_coords[0])/10, head_width=5, color='blue', linestyle='--')
                    # plt.arrow(x_coords[1], y_coords[1], (x_coords[0] - x_coords[1])/10, (y_coords[0] - y_coords[1])/10, head_width=5, color='blue', linestyle='--')

            # Boundary edges.
            # if boundary:
            # x_coords = []
            # y_coords = []
            # for index, edge in enumerate(self.node_edge_generator.graph.edges):
            #     if edge.boundary_edge:
            #         x_coords = (edge.node1.position[0], edge.node2.position[0])
            #         y_coords = (edge.node1.position[1], edge.node2.position[1])
            #     if index == 0 and x_coords != []:
            #         plt.plot(x_coords, y_coords, c="pink", linestyle='--', linewidth=1.5, label="Außenkanten")
            #     elif x_coords != []:
            #         plt.plot(x_coords, y_coords, c="pink", linestyle='--', linewidth=1.5)

        if nodes:
            x = []
            y = []
            # TODO: Is there a smarter way?
            for sample_point in self.node_edge_generator.graph.nodes:
                x.append(sample_point.position[0])
                y.append(sample_point.position[1])
            # x, y = zip(*self.node_edge_generator.graph.nodes)
            nodes_plot = plt.scatter(x, y, c="blue", marker="x", s=40, zorder=2, label="Knoten")

            # x = []
            # y = []
            # for sample_point in self.node_edge_generator.graph.nodes:
            #     if sample_point.node_type == "buffer":
            #         x.append(sample_point.position[0])
            #         y.append(sample_point.position[1])
            # plt.scatter(x, y, c="yellow", marker="X", s=80, label="sample nodes type")

            # Boundary nodes.
            # if boundary:
            x = []
            y = []
            for sample_point in self.node_edge_generator.graph.nodes:
                if sample_point.boundary_node:
                    x.append(sample_point.position[0])
                    y.append(sample_point.position[1])
            # if x != []:
            #     plt.scatter(x, y, c="pink", marker="X", s=80, zorder=2, label="Außenknoten")

            # Convex corner nodes.
            x = []
            y = []
            for sample_point in self.node_edge_generator.graph.nodes:
                if sample_point.corner_type == "convex":
                    x.append(sample_point.position[0])
                    y.append(sample_point.position[1])
            # if x != []:
            #     nodes_convex_plot = plt.scatter(x, y, c="orange", marker="X", s=80, zorder=2, label="Konvexe Eckknoten")

            # Concave corner nodes.
            x = []
            y = []
            for sample_point in self.node_edge_generator.graph.nodes:
                if sample_point.corner_type == "concave":
                    x.append(sample_point.position[0])
                    y.append(sample_point.position[1])
            # if x != []:
            #     nodes_concave_plot = plt.scatter(x, y, c="green", marker="X", s=80, zorder=2, label="Konkave Eckknoten")

            # Buffer nodes.
            x = []
            y = []
            for sample_point in self.node_edge_generator.graph.nodes:
                if sample_point.node_type == "buffer":  # and sample_point.position != (398, 604) and sample_point.position != (462, 604)
                    x.append(sample_point.position[0])
                    y.append(sample_point.position[1])
            if x != []:
                station_buffer_nodes_plot = plt.scatter(x, y, c="yellow", marker="X", s=80, zorder=2, label="Pufferknoten")

        if stations:
            for index, station in enumerate(self.station_config.stations):
                x, y = zip(*station.station_corner_nodes)
                plt.fill(x, y, color='red')
                # x, y = zip(*station.station_corner_nodes, station.station_corner_nodes[0], station.station_corner_nodes[2], station.station_corner_nodes[3], station.station_corner_nodes[1])  # , station.station_corner_nodes[2], station.station_corner_nodes[3], station.station_corner_nodes[1]
                # if index == 0:
                #     station_plot, = plt.plot(x, y, color='red', linestyle='-', label='Stationen')
                # else:
                #     plt.plot(x, y, color='red', linestyle='-')

        if station_nodes:
            # Connection station node to station trajectory node.
            for index, station in enumerate(self.station_config.stations):
                for station_trajectory_node in station.station_trajectory_nodes:
                    x = [station_trajectory_node[0], station.station_node[0]]
                    y = [station_trajectory_node[1], station.station_node[1]]
                    if index == 0:
                        station_trajectory_plot,  = plt.plot(x, y, linewidth=3, color='blue',
                                                             linestyle='--', label='Anfahrtstrajektorien')
                    else:
                        plt.plot(x, y, linewidth=3, color='blue', linestyle='--')

            # Station nodes.
            x, y = zip(*self.station_config.station_nodes)
            station_nodes_plot = plt.scatter(x, y, c="blue", marker="D", s=60, zorder=2, label="Bearbeitungspositionen")

            # Trajectory nodes.
            x, y = zip(*self.station_config.station_trajectory_nodes)
            station_trajectory_nodes_plot = plt.scatter(x, y, c="red", marker="X", s=80, zorder=2, label="Trajekorienknoten")

            # Buffer nodes.
            # x, y = zip(*self.station_config.station_buffer_nodes)
            # plt.scatter(x, y, c="orange", marker="X", s=100, zorder=2, label="Pufferknoten")

        if medial_axis:
            for index, line in enumerate(self.table_config.medial_axis.geoms):
                if index == 0:
                    x, y = line.xy
                    plt.plot(x, y, marker='', color='black', linestyle='--', linewidth=3, label='Mediale Achse')
                else:
                    x, y = line.xy
                    plt.plot(x, y, marker='', color='black', linestyle='--', linewidth=3)

            # x, y = zip(*self.table_config.medial_axis_nodes)
            # plt.scatter(x, y, c="pink", marker="x", label="medial axis nodes")

            # for medial_axis_edge in self.table_config.medial_axis_edges:
            #     x_coords = (medial_axis_edge[0][0], medial_axis_edge[1][0])
            #     y_coords = (medial_axis_edge[0][1], medial_axis_edge[1][1])
            #     plt.plot(x_coords, y_coords, color='pink', linestyle='--')

        if zones:
            # Colorize the zones.
            # colors_in_legend = set()  # Keep track of which colors are in the legend.
            # for index, polygon in enumerate(self.table_config.zones_polygons):
            #     if self.table_config.zones_directions == []:
            #         break
            #     elif self.table_config.zones_directions[index] == 'right':
            #         color = 'red'
            #         label = 'Zone: rechts'
            #     elif self.table_config.zones_directions[index] == 'left':
            #         color = 'blue'
            #         label = 'Zone: links'
            #     elif self.table_config.zones_directions[index] == 'up':
            #         color = 'green'
            #         label = 'Zone: hoch'
            #     elif self.table_config.zones_directions[index] == 'down':
            #         color = 'yellow'
            #         label = 'Zone: runter'
            #     x, y = polygon.exterior.xy
            #     if color not in colors_in_legend:  # Only add to the legend if this color is not already in it.
            #         plt.fill(x, y, color=color, alpha=0.3, label=label)
            #         colors_in_legend.add(color)  # Add this color to the set of colors in the legend.
            #     else:
            #         plt.fill(x, y, color=color, alpha=0.3)  # Don't add a label if this color is already in the legend.

            # Plot the boundary of the zones.
            # for index, line in enumerate(self.table_config.zones_edges.geoms):
            #     if index == 0:
            #         x, y = line.xy
            #         plt.plot(x, y, marker='o', color='blue', linestyle='--', label='zones edges')
            #     else:
            #         x, y = line.xy
            #         plt.plot(x, y, marker='o', color='blue', linestyle='--')

            # Get a list of colors.
            colors = list(mcolors.CSS4_COLORS.keys())

            # Colorize the zones.
            for index, polygon in enumerate(self.table_config.zones_polygons):
                x, y = polygon.exterior.xy
                plt.fill(x, y, color=colors[index + 10], alpha=0.5)

        # self.add_dilation_difference_layer(value=0.7, kernel_radius=self.config.get(Configuration.Boundary_Distance))
        # self.draw_polygon_to_layer("stations_layer", [(100, 100), (122, 100), (122, 122), (100, 122)], value=0.55)
        plt.imshow(self.get_merged_occupancy_layers(layer_ids), cmap="Greys", origin="lower")  # Blues Greys PuBu
        # plt.imshow(self.get_dialated_occupancy_layers(layer_ids, self.config.get(Configuration.Boundary_Distance)), cmap="Greys", origin="lower")  # Blues

        # Create legend objects
        white_patch = mpatches.Patch(color='white', label='Freifläche')
        silver_patch = mpatches.Patch(color='silver', label='Erweiterte Hindernisse')
        grey_patch = mpatches.Patch(color='grey', label='Dauerhafte Hindernisse')
        red_patch = mpatches.Patch(color='red', label='Stationen')
        black_patch = mpatches.Patch(color='black', label='Temporäre Hindernisse')

        # cbar = plt.colorbar(im)
        # cbar.set_label('Occupancy')
        # cbar.set_ticks([0, 1])
        # cbar.set_ticklabels(['White Area Name', 'Black Area Name'])

        # TODO: Plotting with custom colormap.
        # Define the colors for different regions
        # colors = ["white", "blue", "black"]
        # cmap = mcolors.LinearSegmentedColormap.from_list("", colors)

        # Plot the image with the custom colormap
        # plt.imshow(self.get_merged_occupancy_layers(layer_ids), cmap=cmap, origin="lower")

        plt.xlim(0 - 0.5, self.config.get(Configuration.Dim_X) - 0.5)
        plt.ylim(0 - 0.5, self.config.get(Configuration.Dim_Y) - 0.5)

        fontsize = 14
        plt.xlabel('x [cm]', fontsize=fontsize)
        plt.ylabel('y [cm]', fontsize=fontsize)
        plt.tick_params(axis='both', which='major', labelsize=fontsize)

        fontsize = 16
        plt.rc('axes', labelsize=fontsize)
        plt.rc('xtick', labelsize=fontsize)
        plt.rc('ytick', labelsize=fontsize)
        plt.rc('legend', fontsize=fontsize)

        # # handles.append(grey_patch)
        # # handles.append(silver_patch)
        # # handles.append(white_patch)
        # handles.append(modules)
        # # handles.append(table_boundary)
        # # handles.append(cfree)
        # handles.append(red_patch)
        # handles.append(station_trajectory_plot)
        # handles.append(station_nodes_plot)
        # handles.append(station_trajectory_nodes_plot)
        # handles.append(station_buffer_nodes_plot)
        # # handles.append(nodes_convex_plot)
        # # handles.append(nodes_concave_plot)
        # handles.append(nodes_plot)
        # handles.append(edges_plot)
        # handles.append(black_patch)
        # handles.append(grey_patch)
        # handles.append(silver_patch)
        # plt.legend(handles=handles, loc='upper left')  # best

        # plt.legend(loc='best')  # 'center right' 'best
        plt.show(block=block)




from __future__ import annotations

from typing import Generator, TYPE_CHECKING, Optional

import logging

from prodsys.simulation import (
    sim,
    process,
)

from prodsys.simulation.process import (
    ReworkProcess,
)

from prodsys.models.dependency_data import DependencyType
if TYPE_CHECKING:
    from prodsys.simulation import (
        process,
    )
    from prodsys.simulation import request as request_module
    from prodsys.simulation.resources import SystemResource
    from prodsys.simulation.router import Router

logger = logging.getLogger(__name__)

class SystemProcessModelHandler:
    """
    A process model handler is responsible for controlling process models that can contain multiple processes
    with complex dependencies and execution patterns (DAG structures).
    """

    def __init__(self, env: sim.Environment) -> None:
        self.env = env
        self.resource = None

        self.process_model = None
        self.next_possible_processes: Optional[list[process.PROCESS_UNION]] = None
        self.processes_needing_rework: list[process.Process] = []
        self.blocking_rework_process_mappings: list[
            list[process.PROCESS_UNION, list[process.ReworkProcess]]
        ] = []
        self.non_blocking_rework_process_mappings: list[
            list[process.PROCESS_UNION, list[process.ReworkProcess]]
        ] = []
        self.executed_production_processes: list[str] = []

    def get_products_of_lot(
        self, lot_requests: list[request_module.Request]
    ) -> Generator:
        """
        Get the next product for a process. The product is removed (get) from the input queues of the resource.
        """
        for lot_request in lot_requests:
            yield from lot_request.origin_queue.get(lot_request.entity.data.ID)

    def put_entities_of_request(
        self, process_request: request_module.Request
    ) -> Generator:
        """
        Place a product to the output queue (put) of the resource.

        Args:
            process_request (request_module.Request): The request to place the product to.

        Returns:
            Generator: The generator yields when the product is placed in the queue (multiple events for multiple products, e.g. for a batch process or an assembly).
        """
        required_assembly_types = [dependency.data.required_entity for dependency in process_request.required_dependencies if dependency.data.dependency_type == DependencyType.ASSEMBLY]
        for dependant_entity in process_request.entity.depended_entities:
            if dependant_entity.data.type not in required_assembly_types:
                continue    
            dependant_entity.info.log_consumption(dependant_entity.current_locatable, dependant_entity, self.env.now)
        required_tool_types = [dependency.data.required_entity for dependency in process_request.required_dependencies if dependency.data.dependency_type == DependencyType.TOOL]
        for dependant_entity in process_request.entity.depended_entities:
            if dependant_entity.data.type not in required_tool_types:
                continue
            dependant_entity.update_location(process_request.entity._current_locatable)
            dependant_entity.info.log_start_unloading(dependant_entity.current_locatable, dependant_entity, self.env.now, dependant_entity.current_locatable)
            yield from dependant_entity.current_locatable.put(dependant_entity.data)
            dependant_entity.info.log_end_unloading(dependant_entity.current_locatable, dependant_entity, self.env.now, dependant_entity.current_locatable)


    def handle_request(self, process_request: request_module.Request) -> Generator:
        """
        Handle a process model request by executing the processes according to the DAG structure.

        Args:
            process_request (request_module.Request): The process model request.

        Yields:
            Generator: The generator yields when the process model is finished.
        """
        resource: SystemResource = process_request.get_resource()
        self.resource = resource
        proc = process_request.get_process()
        self.process_model = proc.precedence_graph.create_instance()
        
        entity = process_request.get_entity()
        target_queue = process_request.target_queue

        super_system_router = process_request.entity.router
        system_router = resource.router
        process_request.entity.router = system_router

        assert process_request.entity._current_locatable == process_request.origin_queue, f"Product {entity.data.ID} is not at the origin queue {process_request.origin_queue}"
        
        if process_request.required_dependencies:
            yield process_request.request_dependencies()
        
        yield from resource.setup(proc)

        resource.controller.mark_started_process(process_request.capacity_required)
        # TODO: add logging to start and end of process model handling for product entity
        self.set_next_possible_production_processes()
        while self.next_possible_processes:
            executed_process_event = system_router.request_process_step(entity, self.next_possible_processes)
            yield executed_process_event
            # Check if rework is required based on the state's failure determination
            # The state has already determined and logged if the process failed
            if not isinstance(entity.current_process, ReworkProcess):
                # Check if the process failed (determined by the state)
                if entity.last_process_failed:
                    self.add_needed_rework(entity.current_process, system_router)
            elif isinstance(entity.current_process, ReworkProcess):
                self.register_rework(entity.current_process)
            self.update_executed_process(entity.current_process)
            # Reset failure status for next process
            entity.last_process_failed = None
            self.set_next_possible_production_processes()
        
        # After all internal processes are complete, set the entity's current_process to this ProcessModelProcess
        # This ensures that when control returns to a parent handler, it sees the ProcessModelProcess as completed,
        # not the last internal process that was executed
        entity.current_process = proc
        
        if(entity.no_transport_to_sink):
            entity.router.route_disassembled_product_to_sink(entity)
        else:
            arrived_at_queue = system_router.request_transport(entity, target_queue)
            yield arrived_at_queue
        yield from self.put_entities_of_request(process_request)
        process_request.entity.update_location(process_request.target_queue)
        process_request.entity.router = super_system_router
        process_request.entity.router.mark_finished_request(process_request)
        self.resource.controller.mark_finished_process(process_request.capacity_required)


    def add_needed_rework(self, failed_process: process.PROCESS_UNION, router: Router) -> None:
        """
        Adds a process to the list of processes that need rework.

        Args:
            entity: The entity (product) that needs rework.
            failed_process (PROCESS_UNION): Process that needs rework.
        """
        self.processes_needing_rework.append(failed_process)
        possible_rework_processes = router.get_rework_processes(
            failed_process
        )
        if not possible_rework_processes:
            raise ValueError(
                f"No rework processes found for process {failed_process.data.ID}"
            )
        if any(rework_process.blocking for rework_process in possible_rework_processes):
            self.blocking_rework_process_mappings.append(
                [failed_process, possible_rework_processes]
            )
        else:
            self.non_blocking_rework_process_mappings.append(
                [failed_process, possible_rework_processes]
            )

    def register_rework(self, rework_process: process.ReworkProcess) -> None:
        """
        Register a rework process that has been executed.
        
        Args:
            rework_process (process.ReworkProcess): The rework process that has been executed.
        """
        reworked_process = None
        for process_rework_mapping in (
            self.blocking_rework_process_mappings
            + self.non_blocking_rework_process_mappings
        ):
            if rework_process not in process_rework_mapping[1]:
                continue
            reworked_process = process_rework_mapping[0]
            break
        
        if reworked_process is None:
            raise ValueError(f"Rework process {rework_process.data.ID} not found in rework mappings")
            
        if reworked_process in self.processes_needing_rework:
            self.processes_needing_rework.remove(reworked_process)
        
        in_blocking_list = any(
            reworked_process == process_rework_mapping[0]
            for process_rework_mapping in self.blocking_rework_process_mappings
        )
        if in_blocking_list:
            mapping_to_adjust = self.blocking_rework_process_mappings
        else:
            mapping_to_adjust = self.non_blocking_rework_process_mappings
        
        index_to_remove = None
        for index, process_rework_mapping in enumerate(mapping_to_adjust):
            if reworked_process == process_rework_mapping[0]:
                index_to_remove = index
                break
        
        if index_to_remove is None:
            raise ValueError(f"Rework process {rework_process.data.ID} not found in rework mappings")
        mapping_to_adjust.pop(index_to_remove)

    def update_executed_process(self, executed_process: process.PROCESS_UNION) -> None:
        if not isinstance(executed_process, process.ReworkProcess):
            self.process_model.update_marking_from_transition(executed_process)  # type: ignore
        self.executed_production_processes.append(executed_process.data.ID)

    def set_next_possible_production_processes(self):
        """
        Sets the next process of the product object based on the current state of the product and its process model.
        """
        # check if rework is needed due to blocking rework processes. If so, execute these rework processes
        if self.blocking_rework_process_mappings:
            next_possible_processes = []
            for process_mapping in self.blocking_rework_process_mappings:
                possible_rework_processes = process_mapping[1]
                next_possible_processes += possible_rework_processes
            self.next_possible_processes = next_possible_processes
            return

        next_possible_processes = self.process_model.get_next_possible_processes()

        # if all normal processes are done, i.e. the product has finished its process sequence, execute rework processes without blocking.
        if not next_possible_processes and self.non_blocking_rework_process_mappings:
            next_possible_processes = []
            for process_mapping in self.non_blocking_rework_process_mappings:
                possible_rework_processes = process_mapping[1]
                next_possible_processes += possible_rework_processes
            next_possible_processes = next_possible_processes
            self.next_possible_processes = next_possible_processes
            return
        # if all normal processes are done and no rework processes are needed, the product has finished its process sequence.
        if not next_possible_processes:
            self.next_possible_processes = None
            return

        self.next_possible_processes = next_possible_processes


from prodsys.models.production_system_data import ProductionSystemData
from prodsys.models.scenario_data import ReconfigurationEnum
from prodsys.optimization.adapter_manipulation import add_transformation_operation
from prodsys.optimization.math_opt import MathOptHyperparameters
from prodsys.optimization.optimizer import FileSystemSaveOptimizer, Optimizer
import time
import logging

logger = logging.getLogger(__name__)


def main():
    # Setzen der Hyperparameter für mathematische Optimierung
    hyper_parameters = MathOptHyperparameters(
        full_save="data/math_results",
        optimization_time_portion=1,
        number_of_solutions=1,
        adjusted_number_of_transport_resources=2,
        number_of_seeds=2,
    )

    def new_transformation(adapter: ProductionSystemData) -> bool:
        print("Transformation function called.")

    add_transformation_operation(
        transformation=ReconfigurationEnum.PRODUCTION_CAPACITY,
        operation=new_transformation,
    )

    base_configuration = ProductionSystemData.read(
        "examples/optimization/optimization_example/base_scenario.json",
    )
    base_configuration.read_scenario(
        "examples/optimization/optimization_example/scenario.json",
    )

    optimizer = FileSystemSaveOptimizer(
        adapter=base_configuration,
        hyperparameters=hyper_parameters,
        save_folder="data/math_results",
    )

    optimizer.optimize()


if __name__ == "__main__":
    main()
